import time


class EvaluationsScoreTracker:
    def __init__(
        self, max_evaluations: int, stop_at_score: float, print_period: int = 10
    ):
        self.max_evaluations = max_evaluations
        self.stop_at_score = stop_at_score
        self.evaluations = 0
        self.score = float("inf")
        self.history = [(0, float("inf"))]
        self._scale_by = 1
        self._scale_score_by = 1
        self.print_period = print_period
        self.prev_print_time = None
        self.print_stopped_text = False
        self.n_mappings = {}
        self.runtime = {}

    def add_evaluation(self, n_evaluations: int, best_score: float):
        self.evaluations += n_evaluations * self._scale_by
        new_score = best_score * self._scale_score_by
        if new_score < self.score:
            print(f"New score {new_score} after evaluation {self.evaluations}")
        self.score = min(self.score, best_score * self._scale_score_by)
        # Same score as before, remove the last entry
        if len(self.history) > 2 and self.history[-2][1] == self.score:
            self.history.pop(-1)
        self.history.append((self.evaluations, self.score))

        cur_time = time.time()
        if (
            self.prev_print_time is None
            or cur_time - self.prev_print_time > self.print_period
        ):
            self.prev_print_time = cur_time

        if self.max_evaluations is not None and self.evaluations > self.max_evaluations:
            self.clean_history()
            if not self.print_stopped_text:
                print(
                    f"Stopping due to evaluations {self.evaluations} > {self.max_evaluations}"
                )
                self.print_stopped_text = True
            return True
        if self.stop_at_score is not None and self.score < self.stop_at_score:
            self.clean_history()
            if not self.print_stopped_text:
                print(f"Stopping due to score {self.score} < {self.stop_at_score}")
                self.print_stopped_text = True
            return True
        return False

    def finished(self):
        enough_evaluations = (
            self.max_evaluations is not None and self.evaluations > self.max_evaluations
        )
        enough_score = (
            self.stop_at_score is not None and self.score < self.stop_at_score
        )
        return enough_evaluations or enough_score

    def multiply_score_by(self, scale_by: float):
        self._scale_score_by *= scale_by

    def __repr__(self):
        return f"Evaluations: {self.evaluations}, Score: {self.score}"

    def __str__(self):
        return f"Evaluations: {self.evaluations}, Score: {self.score}"

    def clean_history(self):
        keep_indices = [0]
        for i in range(1, len(self.history) - 1):
            if (
                self.history[i][1] != self.history[i - 1][1]
                or self.history[i][1] != self.history[i + 1][1]
            ):
                keep_indices.append(i)
        keep_indices.append(len(self.history) - 1)
        self.history = [self.history[i] for i in keep_indices]

    def merge_with(self, other: "EvaluationsScoreTracker"):
        self.score = min(self.score, other.score)
        self.evaluations += other.evaluations

        i, j = 1, 1
        history = [(0, float("inf"))]
        cur_score = float("inf")
        cur_evaluations = 0
        while i < len(self.history) or j < len(other.history):
            # Grab whichever has the lowest evaluations
            if i < len(self.history) and (
                j == len(other.history) or self.history[i][0] < other.history[j][0]
            ):
                new_evaluations = self.history[i][0] - self.history[i - 1][0]
                new_score = self.history[i][1]
                cur_evaluations += new_evaluations
                cur_score = min(cur_score, new_score)
                history.append((cur_evaluations, cur_score))
                i += 1
            elif j < len(other.history):
                new_evaluations = other.history[j][0] - other.history[j - 1][0]
                new_score = other.history[j][1]
                cur_evaluations += new_evaluations
                cur_score = min(cur_score, new_score)
                history.append((cur_evaluations, cur_score))
                j += 1
        self.history = history
        self.clean_history()

    def increase_all_evaluations(self, n_evaluations: int):
        self.evaluations += n_evaluations
        self.history = [(e + n_evaluations, s) for e, s in self.history]



import pickle
from pathlib import Path
from typing import Callable, TypeVar


T = TypeVar("T")


class PickleCache:
    def __init__(self, fname: Path):
        self.fname = fname

    def get(self, cache_miss_thunk: Callable[[], T]) -> T:
        """
        Return loaded pickle at `self.fname` if exists;
        otherwise, calls `cache_miss_thunk` and stores result in `self.fname`.
        """
        print(self.fname)
        if self.fname.exists():
            with open(self.fname, "rb") as f:
                return pickle.load(f)
        else:
            result = cache_miss_thunk()
            with open(self.fname, "wb") as f:
                pickle.dump(result, f)
            return result

    def set(self, data: T) -> None:
        """Set data at `self.fname`."""
        with open(self.fname, "wb") as f:
            pickle.dump(data, f)
        return data



"""
Tests for expression evaluation:
  - Spec._spec_eval_expressions (variables, renames, workload, arch)
  - Workload bits_per_value resolution
  - Workload persistent_tensors resolution
  - Arch evaluation with variables and expressions
  - Partial evaluation (eval_arch only, eval_non_arch only)
  - Variables referencing other variables
  - EvaluationError for bad expressions
"""

import unittest
from pathlib import Path

from accelforge.frontend.spec import Spec
from accelforge.frontend.workload import Workload, Einsum
from accelforge.frontend.arch import Arch, Memory, Compute, Fanout, Toll
from accelforge.frontend.arch import Spatial as ArchSpatial
from accelforge.frontend.renames import Renames, EinsumRename, Rename
from accelforge.frontend.variables import Variables
from accelforge.frontend.config import Config

_REPO_ROOT = Path(__file__).parent.parent.parent
EXAMPLES_DIR = _REPO_ROOT / "examples"


# ============================================================================
# Spec eval basic
# ============================================================================


class TestSpecEvalExpressions(unittest.TestCase):
    """Test that _spec_eval_expressions evaluates the spec correctly."""

    def _simple_spec(self, **overrides):
        """A simple spec with one matmul einsum."""
        data = dict(
            workload=Workload(
                rank_sizes={"M": 64, "K": 32, "N": 64},
                bits_per_value={"All": 8},
                einsums=[
                    {
                        "name": "Matmul",
                        "tensor_accesses": [
                            {"name": "A", "projection": ["m", "k"]},
                            {"name": "B", "projection": ["k", "n"]},
                            {"name": "C", "projection": ["m", "n"], "output": True},
                        ],
                    }
                ],
            ),
            arch=Arch(
                nodes=[
                    Memory(
                        name="MainMemory",
                        size=1_000_000,
                        actions=[
                            {"name": "read", "energy": 1, "latency": 0},
                            {"name": "write", "energy": 1, "latency": 0},
                        ],
                        leak_power=0,
                        area=0,
                    ),
                    Compute(
                        name="MAC",
                        actions=[{"name": "compute", "energy": 1, "latency": 1}],
                        leak_power=0,
                        area=0,
                    ),
                ]
            ),
        )
        data.update(overrides)
        return Spec(**data)

    def test_eval_sets_evaluated_flag(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions()
        self.assertTrue(getattr(evaluated, "_evaluated", False))

    def test_eval_resolves_bits_per_value(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions()
        for ta in evaluated.workload.einsums[0].tensor_accesses:
            self.assertIsNotNone(ta.bits_per_value)
            self.assertEqual(ta.bits_per_value, 8)

    def test_eval_preserves_rank_sizes(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions()
        self.assertEqual(evaluated.workload.rank_sizes["M"], 64)
        self.assertEqual(evaluated.workload.rank_sizes["K"], 32)

    def test_eval_preserves_arch(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions()
        mem = evaluated.arch.find("MainMemory")
        self.assertEqual(mem.name, "MainMemory")

    def test_eval_with_einsum_name(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions(einsum_name="Matmul")
        self.assertTrue(getattr(evaluated, "_evaluated", False))

    def test_eval_arch_only(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions(eval_arch=True, eval_non_arch=False)
        self.assertIsNotNone(evaluated)

    def test_eval_non_arch_only(self):
        spec = self._simple_spec()
        evaluated = spec._spec_eval_expressions(eval_arch=False, eval_non_arch=True)
        self.assertIsNotNone(evaluated)

    def test_eval_with_variables(self):
        spec = self._simple_spec(variables=Variables(my_size=256))
        evaluated = spec._spec_eval_expressions()
        self.assertIsNotNone(evaluated)

    def test_calling_eval_twice_idempotent(self):
        spec = self._simple_spec()
        evaluated1 = spec._spec_eval_expressions()
        evaluated2 = evaluated1._spec_eval_expressions()
        self.assertTrue(getattr(evaluated2, "_evaluated", False))


# ============================================================================
# Bits per value evaluation
# ============================================================================


class TestBitsPerValueEvaluation(unittest.TestCase):
    """Test bits_per_value resolution during evaluation."""

    def test_all_tensors_get_bits(self):
        spec = Spec(
            workload=Workload(
                rank_sizes={"M": 16, "N": 16},
                bits_per_value={"All": 16},
                einsums=[
                    {
                        "name": "E",
                        "tensor_accesses": [
                            {"name": "X", "projection": ["m"]},
                            {"name": "Y", "projection": ["m", "n"], "output": True},
                        ],
                    }
                ],
            ),
        )
        evaluated = spec._spec_eval_expressions()
        for ta in evaluated.workload.einsums[0].tensor_accesses:
            self.assertEqual(ta.bits_per_value, 16)

    def test_per_tensor_bits_override(self):
        """Per-tensor-access bits_per_value overrides the workload-level one."""
        spec = Spec(
            workload=Workload(
                rank_sizes={"M": 16, "N": 16},
                bits_per_value={"All": 8},
                einsums=[
                    {
                        "name": "E",
                        "tensor_accesses": [
                            {"name": "X", "projection": ["m"], "bits_per_value": 32},
                            {"name": "Y", "projection": ["m", "n"], "output": True},
                        ],
                    }
                ],
            ),
        )
        evaluated = spec._spec_eval_expressions()
        x_ta = [
            t for t in evaluated.workload.einsums[0].tensor_accesses if t.name == "X"
        ][0]
        y_ta = [
            t for t in evaluated.workload.einsums[0].tensor_accesses if t.name == "Y"
        ][0]
        self.assertEqual(x_ta.bits_per_value, 32)
        self.assertEqual(y_ta.bits_per_value, 8)


# ============================================================================
# Persistent tensors evaluation
# ============================================================================


class TestPersistentTensorsEvaluation(unittest.TestCase):
    """Test persistent_tensors field evaluation."""

    def test_persistent_tensors_marked(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B_concise.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path)
        evaluated = spec._spec_eval_expressions()

        weight_tensors = {"WV", "WK", "WQ", "WZ", "WFFA", "WFFB"}
        for einsum in evaluated.workload.einsums:
            for ta in einsum.tensor_accesses:
                if ta.name in weight_tensors:
                    self.assertTrue(
                        ta.persistent,
                        f"{ta.name} in {einsum.name} should be persistent",
                    )

    def test_intermediate_tensors_not_persistent(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B_concise.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path)
        evaluated = spec._spec_eval_expressions()

        # I is an intermediate. It should not be persistent.
        v_einsum = [e for e in evaluated.workload.einsums if e.name == "V"][0]
        i_ta = [t for t in v_einsum.tensor_accesses if t.name == "I"][0]
        self.assertFalse(i_ta.persistent)


# ============================================================================
# Renames evaluation
# ============================================================================


class TestRenamesEvaluation(unittest.TestCase):
    """Test that renames are properly resolved during spec evaluation."""

    def test_default_renames_applied(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path)
        evaluated = spec._spec_eval_expressions()

        # Check that the default renames (input, output, weight) are applied
        for einsum in evaluated.workload.einsums:
            rename_names = {r.name for r in einsum.renames}
            self.assertIn("input", rename_names)
            self.assertIn("output", rename_names)
            self.assertIn("weight", rename_names)

    def test_inline_renames_applied(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path)
        evaluated = spec._spec_eval_expressions()

        # Matmul1 has an inline rename: input: T0
        matmul1 = evaluated.workload.einsums["Matmul1"]
        # The inline rename should set the "input" rename to T0
        input_rename = [r for r in matmul1.renames if r.name == "input"]
        self.assertGreater(len(input_rename), 0)

    def test_builtin_renames_available(self):
        """Built-in renames like All, Inputs, Outputs, Nothing are always available."""
        yaml_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path)
        evaluated = spec._spec_eval_expressions()

        matmul1 = evaluated.workload.einsums["Matmul1"]
        rename_names = {r.name for r in matmul1.renames}
        for builtin in ["All", "Tensors", "Nothing", "Inputs", "Outputs"]:
            self.assertIn(builtin, rename_names)


# ============================================================================
# YAML evaluation with Jinja
# ============================================================================


class TestJinjaEvaluation(unittest.TestCase):
    """Test that Jinja2 templating works during spec evaluation."""

    def test_jinja_variable_substitution(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "matmuls.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(
            yaml_path, jinja_parse_data={"N_EINSUMS": 2, "M": 128, "KN": 64}
        )
        self.assertEqual(len(spec.workload.einsums), 2)

    def test_jinja_three_einsums(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "matmuls.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path, jinja_parse_data={"N_EINSUMS": 3})
        self.assertEqual(len(spec.workload.einsums), 3)

    def test_jinja_default_values(self):
        """GPT concise uses Jinja defaults for BATCH_SIZE and N_TOKENS."""
        yaml_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B_concise.yaml"
        if not yaml_path.exists():
            self.skipTest(f"YAML file not found: {yaml_path}")
        spec = Spec.from_yaml(yaml_path)
        # Default BATCH_SIZE is 1, N_TOKENS is 8192
        self.assertEqual(spec.workload.rank_sizes["B"], 1)
        self.assertEqual(spec.workload.rank_sizes["M"], 8192)


# ============================================================================
# Multi-file spec evaluation
# ============================================================================


class TestMultiFileSpecEval(unittest.TestCase):
    """Test spec evaluation when loading multiple YAML files."""

    def test_load_arch_and_workload(self):
        arch_path = EXAMPLES_DIR / "arches" / "simple.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "matmuls.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path, jinja_parse_data={"N_EINSUMS": 1})
        self.assertGreater(len(spec.workload.einsums), 0)
        self.assertGreater(len(spec.arch.nodes), 0)

    def test_full_spec_evaluates(self):
        arch_path = EXAMPLES_DIR / "arches" / "simple.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "matmuls.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path, jinja_parse_data={"N_EINSUMS": 1})
        evaluated = spec._spec_eval_expressions()
        self.assertTrue(getattr(evaluated, "_evaluated", False))

    def test_full_spec_with_mapping(self):
        arch_path = EXAMPLES_DIR / "arches" / "simple.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "matmuls.yaml"
        map_path = EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml"
        if not all(p.exists() for p in [arch_path, wl_path, map_path]):
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(
            arch_path,
            wl_path,
            map_path,
            jinja_parse_data={"N_EINSUMS": 1, "M": 64, "KN": 32},
        )
        evaluated = spec._spec_eval_expressions()
        self.assertTrue(getattr(evaluated, "_evaluated", False))
        # Mapping should have nodes
        self.assertGreater(len(evaluated.mapping.nodes), 0)


# ============================================================================
# Evaluation with arch expressions
# ============================================================================


class TestArchExpressionEvaluation(unittest.TestCase):
    """Test that arch expressions (e.g., size: 1024*1024*128*8) are evaluated."""

    def test_arch_size_expression(self):
        """Arch memory sizes that are expressions get evaluated.
        TPU arch needs einsum context for some expressions (e.g., len(All))."""
        arch_path = EXAMPLES_DIR / "arches" / "tpu_v4i.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path)
        # Evaluate with einsum context to handle `weight.bits_per_value` etc.
        evaluated = spec._spec_eval_expressions(einsum_name="Matmul1")
        glb = evaluated.arch.find("GlobalBuffer")
        # 1024*1024*128*8 = 1073741824
        self.assertEqual(glb.size, 1024 * 1024 * 128 * 8)

    def test_arch_latency_expression(self):
        """Arch action latency expressions like 1/(8*614e9) get evaluated."""
        arch_path = EXAMPLES_DIR / "arches" / "tpu_v4i.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path)
        # Evaluate with einsum context
        evaluated = spec._spec_eval_expressions(einsum_name="Matmul1")
        mm = evaluated.arch.find("MainMemory")
        read_action = [a for a in mm.actions if a.name == "read"][0]
        self.assertIsNotNone(read_action.latency)
        self.assertAlmostEqual(read_action.latency, 1 / (8 * 614e9), places=20)


# ============================================================================
# Concise and verbose equivalence
# ============================================================================


class TestConciseVerboseEquivalence(unittest.TestCase):
    """Check that concise and verbose workloads produce the same evaluated results."""

    def test_same_einsum_names(self):
        concise_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B_concise.yaml"
        verbose_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B.yaml"
        if not concise_path.exists() or not verbose_path.exists():
            self.skipTest("YAML files not found")

        c = Spec.from_yaml(concise_path)._spec_eval_expressions()
        v = Spec.from_yaml(verbose_path)._spec_eval_expressions()
        self.assertEqual(
            {e.name for e in c.workload.einsums},
            {e.name for e in v.workload.einsums},
        )

    def test_same_tensor_names_per_einsum(self):
        concise_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B_concise.yaml"
        verbose_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B.yaml"
        if not concise_path.exists() or not verbose_path.exists():
            self.skipTest("YAML files not found")

        c = Spec.from_yaml(concise_path)._spec_eval_expressions()
        v = Spec.from_yaml(verbose_path)._spec_eval_expressions()
        for c_einsum in c.workload.einsums:
            v_einsum = v.workload.einsums[c_einsum.name]
            self.assertEqual(
                c_einsum.tensor_names,
                v_einsum.tensor_names,
                f"Tensor names differ for {c_einsum.name}",
            )

    def test_same_bits_per_value(self):
        concise_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B_concise.yaml"
        verbose_path = EXAMPLES_DIR / "workloads" / "gpt3_6.7B.yaml"
        if not concise_path.exists() or not verbose_path.exists():
            self.skipTest("YAML files not found")

        c = Spec.from_yaml(concise_path)._spec_eval_expressions()
        v = Spec.from_yaml(verbose_path)._spec_eval_expressions()
        for c_einsum in c.workload.einsums:
            v_einsum = v.workload.einsums[c_einsum.name]
            for c_ta in c_einsum.tensor_accesses:
                v_ta = v_einsum.tensor_accesses[c_ta.name]
                self.assertEqual(
                    c_ta.bits_per_value,
                    v_ta.bits_per_value,
                    f"bits_per_value differs for {c_ta.name} in {c_einsum.name}",
                )


if __name__ == "__main__":
    unittest.main()



"""
Tests for:
  - Config
  - Variables
  - Model
  - FFM (Fast and Fusiest Mapper config)
  - Metrics
"""

import unittest
import math

from accelforge.frontend.config import Config
from accelforge.frontend.variables import Variables
from accelforge.frontend.model import Model
from accelforge.frontend.mapper.ffm import FFM
from accelforge.frontend.mapper.metrics import Metrics


# ============================================================================
# Config
# ============================================================================


class TestConfig(unittest.TestCase):
    """Tests for the Config model."""

    def test_default_creation(self):
        c = Config()
        self.assertEqual(len(c.expression_custom_functions), 0)
        self.assertEqual(len(c.component_models), 0)

    def test_use_installed_component_models_default(self):
        c = Config()
        self.assertTrue(c.use_installed_component_models)

    def test_use_installed_component_models_false(self):
        c = Config(use_installed_component_models=False)
        self.assertFalse(c.use_installed_component_models)

    def test_custom_functions_list(self):
        c = Config(expression_custom_functions=["path/to/file.py"])
        self.assertEqual(len(c.expression_custom_functions), 1)

    def test_component_models_list(self):
        c = Config(component_models=["path/to/model.py"])
        self.assertEqual(len(c.component_models), 1)


# ============================================================================
# Variables
# ============================================================================


class TestVariables(unittest.TestCase):
    """Tests for the Variables model."""

    def test_empty_creation(self):
        v = Variables()
        self.assertIsNotNone(v)

    def test_custom_attributes(self):
        """Variables supports arbitrary extra attributes via EvalExtras."""
        v = Variables(a=123, b="a + 5")
        self.assertEqual(v.a, 123)
        self.assertEqual(v.b, "a + 5")


# ============================================================================
# Model
# ============================================================================


class TestModel(unittest.TestCase):
    """Tests for the Model configuration."""

    def test_default_creation(self):
        m = Model()
        self.assertIsNotNone(m.metrics)

    def test_default_metrics_all(self):
        m = Model()
        self.assertEqual(m.metrics, Metrics.all_metrics())


# ============================================================================
# Metrics
# ============================================================================


class TestMetrics(unittest.TestCase):
    """Tests for the Metrics flag enum."""

    def test_latency_exists(self):
        self.assertIsNotNone(Metrics.LATENCY)

    def test_energy_exists(self):
        self.assertIsNotNone(Metrics.ENERGY)

    def test_resource_usage_exists(self):
        self.assertIsNotNone(Metrics.RESOURCE_USAGE)

    def test_actions_exists(self):
        self.assertIsNotNone(Metrics.ACTIONS)

    def test_all_metrics_includes_all(self):
        all_m = Metrics.all_metrics()
        self.assertIn(Metrics.LATENCY, all_m)
        self.assertIn(Metrics.ENERGY, all_m)
        self.assertIn(Metrics.RESOURCE_USAGE, all_m)
        self.assertIn(Metrics.ACTIONS, all_m)

    def test_combining_metrics(self):
        combined = Metrics.LATENCY | Metrics.ENERGY
        self.assertIn(Metrics.LATENCY, combined)
        self.assertIn(Metrics.ENERGY, combined)
        self.assertNotIn(Metrics.RESOURCE_USAGE, combined)

    def test_individual_metric_values_are_unique(self):
        vals = [
            Metrics.LATENCY,
            Metrics.ENERGY,
            Metrics.RESOURCE_USAGE,
            Metrics.ACTIONS,
        ]
        self.assertEqual(len(vals), len(set(vals)))


# ============================================================================
# FFM
# ============================================================================


class TestFFM(unittest.TestCase):
    """Tests for the FFM mapper configuration."""

    def test_default_creation(self):
        f = FFM()
        self.assertIsNotNone(f)

    def test_default_metrics(self):
        f = FFM()
        self.assertEqual(f.metrics, Metrics.ENERGY)

    def test_default_info_metrics(self):
        f = FFM()
        self.assertEqual(f.info_metrics, Metrics.all_metrics())

    def test_force_memory_hierarchy_order_default(self):
        f = FFM()
        self.assertTrue(f.force_memory_hierarchy_order)

    def test_max_fused_loops_per_rank_variable(self):
        f = FFM()
        self.assertEqual(f.max_fused_loops_per_rank_variable, 1)

    def test_max_fused_loops_default_inf(self):
        f = FFM()
        self.assertEqual(f.max_fused_loops, float("inf"))

    def test_max_loops_default_inf(self):
        f = FFM()
        self.assertEqual(f.max_loops, float("inf"))

    def test_max_loops_minus_ranks_default_inf(self):
        f = FFM()
        self.assertEqual(f.max_loops_minus_ranks, float("inf"))

    def test_memory_limit_default_inf(self):
        f = FFM()
        self.assertEqual(f.memory_limit, float("inf"))

    def test_memory_limit_per_process_default_inf(self):
        f = FFM()
        self.assertEqual(f.memory_limit_per_process, float("inf"))

    def test_time_limit_default_inf(self):
        f = FFM()
        self.assertEqual(f.time_limit, float("inf"))

    def test_time_limit_per_pmapping_template_default_inf(self):
        f = FFM()
        self.assertEqual(f.time_limit_per_pmapping_template, float("inf"))

    def test_max_pmapping_templates_default_inf(self):
        f = FFM()
        self.assertEqual(f.max_pmapping_templates_per_einsum, float("inf"))

    def test_custom_metrics(self):
        f = FFM(metrics=Metrics.LATENCY)
        self.assertEqual(f.metrics, Metrics.LATENCY)

    def test_combined_metrics(self):
        f = FFM(metrics=Metrics.LATENCY | Metrics.ENERGY)
        self.assertIn(Metrics.LATENCY, f.metrics)
        self.assertIn(Metrics.ENERGY, f.metrics)

    def test_out_of_order_default_false(self):
        f = FFM()
        self.assertFalse(
            f.out_of_order_hierarchy_explore_removing_spatials_for_more_temporals
        )

    def test_custom_max_fused_loops(self):
        f = FFM(max_fused_loops=3)
        self.assertEqual(f.max_fused_loops, 3)


if __name__ == "__main__":
    unittest.main()



"""
Tests for mapping frontend classes:
  - MappingNode, MappingNodeWithChildren
  - Temporal, Spatial (mapping loops)
  - Storage, Toll (mapping tensor holders)
  - Compute (mapping compute)
  - TilePattern
  - Nested, Split, Sequential, Pipeline
  - Reservation
  - Mapping (top-level)
"""

import unittest

from accelforge.frontend.mapping.mapping import (
    Compute,
    Loop,
    Mapping,
    MappingNode,
    MappingNodeWithChildren,
    Nested,
    Pipeline,
    Reservation,
    Sequential,
    Spatial,
    Split,
    Storage,
    Temporal,
    TensorHolder,
    TilePattern,
    Toll,
)


# ============================================================================
# TilePattern
# ============================================================================


class TestTilePattern(unittest.TestCase):
    """Tests for TilePattern dataclass."""

    def test_default_creation(self):
        tp = TilePattern()
        self.assertEqual(tp.tile_shape, "symbol")
        self.assertEqual(tp.initial_tile_shape, "symbol")
        self.assertIsNone(tp.calculated_n_iterations)

    def test_with_values(self):
        tp = TilePattern(tile_shape=4, initial_tile_shape=8)
        self.assertEqual(tp.tile_shape, 4)
        self.assertEqual(tp.initial_tile_shape, 8)

    def test_str_with_iterations(self):
        tp = TilePattern(tile_shape=4, initial_tile_shape=8, calculated_n_iterations=3)
        s = str(tp)
        self.assertIn("3", s)
        self.assertIn("4", s)
        self.assertIn("8", s)

    def test_str_symbol_only(self):
        tp = TilePattern()
        s = str(tp)
        # Should be empty since all are "symbol"
        self.assertEqual(s, "")

    def test_update(self):
        tp = TilePattern(tile_shape=4)
        tp2 = tp.update(tile_shape=8)
        self.assertEqual(tp2.tile_shape, 8)
        self.assertEqual(tp.tile_shape, 4)  # Original unchanged

    def test_equality(self):
        tp1 = TilePattern(tile_shape=4, initial_tile_shape=8)
        tp2 = TilePattern(tile_shape=4, initial_tile_shape=8)
        self.assertEqual(tp1, tp2)

    def test_inequality(self):
        tp1 = TilePattern(tile_shape=4)
        tp2 = TilePattern(tile_shape=8)
        self.assertNotEqual(tp1, tp2)

    def test_hash(self):
        tp1 = TilePattern(tile_shape=4, initial_tile_shape=8)
        tp2 = TilePattern(tile_shape=4, initial_tile_shape=8)
        self.assertEqual(hash(tp1), hash(tp2))

    def test_clear_symbols(self):
        tp = TilePattern(tile_shape="x", initial_tile_shape="y")
        cleared = tp._clear_symbols()
        self.assertIsNone(cleared.tile_shape)
        self.assertIsNone(cleared.initial_tile_shape)

    def test_clear_symbols_keeps_ints(self):
        tp = TilePattern(tile_shape=4, initial_tile_shape=8)
        cleared = tp._clear_symbols()
        self.assertEqual(cleared.tile_shape, 4)
        self.assertEqual(cleared.initial_tile_shape, 8)

    def test_as_str_selective(self):
        tp = TilePattern(tile_shape=4, initial_tile_shape=8, calculated_n_iterations=2)
        s = tp.as_str(with_initial_tile_shape=False, with_tile_shape=True)
        self.assertNotIn("initial=", s)
        self.assertIn("tile_shape=4", s)

    def test_symbol_attrs(self):
        tp = TilePattern()
        attrs = tp._symbol_attrs()
        self.assertIn("tile_shape", attrs)
        self.assertIn("initial_tile_shape", attrs)
        self.assertIn("calculated_n_iterations", attrs)


# ============================================================================
# Temporal
# ============================================================================


class TestTemporal(unittest.TestCase):
    """Tests for Temporal loop node."""

    def test_basic_creation(self):
        t = Temporal(rank_variable="m", tile_shape=4)
        self.assertEqual(t.rank_variable, "m")
        self.assertEqual(t.tile_shape, 4)

    def test_str_representation(self):
        t = Temporal(rank_variable="m", tile_shape=4)
        s = str(t)
        self.assertIn("for", s)
        self.assertIn("m", s)

    def test_compact_str(self):
        t = Temporal(rank_variable="m", tile_shape=4)
        s = t.compact_str()
        self.assertIn("T-", s)
        self.assertIn("m", s)

    def test_tile_pattern(self):
        t = Temporal(rank_variable="m", tile_shape=4, initial_tile_shape=8)
        tp = t.tile_pattern
        self.assertEqual(tp.tile_shape, 4)
        self.assertEqual(tp.initial_tile_shape, 8)

    def test_equality(self):
        t1 = Temporal(rank_variable="m", tile_shape=4)
        t2 = Temporal(rank_variable="m", tile_shape=4)
        self.assertEqual(t1, t2)

    def test_inequality_different_rv(self):
        t1 = Temporal(rank_variable="m", tile_shape=4)
        t2 = Temporal(rank_variable="n", tile_shape=4)
        self.assertNotEqual(t1, t2)

    def test_inequality_different_tile(self):
        t1 = Temporal(rank_variable="m", tile_shape=4)
        t2 = Temporal(rank_variable="m", tile_shape=8)
        self.assertNotEqual(t1, t2)

    def test_merge_different_tile_raises(self):
        t1 = Temporal(rank_variable="m", tile_shape=4)
        t2 = Temporal(rank_variable="n", tile_shape=8)
        with self.assertRaises(ValueError):
            t1._merge(t2)

    def test_merge_non_temporal_raises(self):
        t1 = Temporal(rank_variable="m", tile_shape=4)
        s1 = Spatial(rank_variable="n", tile_shape=4, name="X", component="Mem")
        with self.assertRaises(ValueError):
            t1._merge(s1)


# ============================================================================
# Spatial (mapping)
# ============================================================================


class TestMappingSpatial(unittest.TestCase):
    """Tests for Spatial loop node in mapping."""

    def test_basic_creation(self):
        s = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        self.assertEqual(s.rank_variable, "m")
        self.assertEqual(s.name, "X")
        self.assertEqual(s.component, "Register")

    def test_str_representation(self):
        s = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        result = str(s)
        self.assertIn("S-X", result)
        self.assertIn("m", result)

    def test_compact_str(self):
        s = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        result = s.compact_str()
        self.assertIn("S-X", result)
        self.assertIn("m", result)

    def test_equality(self):
        s1 = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        s2 = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        self.assertEqual(s1, s2)

    def test_inequality_different_name(self):
        s1 = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        s2 = Spatial(rank_variable="m", tile_shape=4, name="Y", component="Register")
        self.assertNotEqual(s1, s2)

    def test_inequality_different_component(self):
        s1 = Spatial(rank_variable="m", tile_shape=4, name="X", component="Register")
        s2 = Spatial(rank_variable="m", tile_shape=4, name="X", component="Buffer")
        self.assertNotEqual(s1, s2)

    def test_merge_different_name_raises(self):
        s1 = Spatial(rank_variable="m", tile_shape=4, name="X", component="Reg")
        s2 = Spatial(rank_variable="n", tile_shape=4, name="Y", component="Reg")
        with self.assertRaises(ValueError):
            s1._merge(s2)


# ============================================================================
# Storage
# ============================================================================


class TestStorage(unittest.TestCase):
    """Tests for Storage mapping node."""

    def test_basic_creation(self):
        s = Storage(tensors=["A", "B"], component="MainMem")
        self.assertEqual(s.tensors, ["A", "B"])
        self.assertEqual(s.component, "MainMem")

    def test_str_representation(self):
        s = Storage(tensors=["A"], component="MainMem")
        result = str(s)
        self.assertIn("MainMem", result)
        self.assertIn("A", result)

    def test_compact_str(self):
        s = Storage(tensors=["A"], component="MainMem")
        result = s.compact_str()
        self.assertIn("A", result)
        self.assertIn("MainMem", result)

    def test_tensor_property_single(self):
        s = Storage(tensors=["A"], component="MainMem")
        self.assertEqual(s.tensor, "A")

    def test_tensor_property_multiple_raises(self):
        s = Storage(tensors=["A", "B"], component="MainMem")
        with self.assertRaises(ValueError):
            _ = s.tensor

    def test_equality(self):
        s1 = Storage(tensors=["A", "B"], component="MainMem")
        s2 = Storage(tensors=["A", "B"], component="MainMem")
        self.assertEqual(s1, s2)

    def test_equality_order_matters_for_lists(self):
        s1 = Storage(tensors=["A", "B"], component="MainMem")
        s2 = Storage(tensors=["B", "A"], component="MainMem")
        # set comparison in __eq__
        self.assertEqual(s1, s2)

    def test_inequality_different_component(self):
        s1 = Storage(tensors=["A"], component="MainMem")
        s2 = Storage(tensors=["A"], component="Buffer")
        self.assertNotEqual(s1, s2)

    def test_merge_different_component_raises(self):
        s1 = Storage(tensors=["A"], component="MainMem")
        s2 = Storage(tensors=["B"], component="Buffer")
        with self.assertRaises(ValueError):
            s1._merge(s2)

    def test_render_node_shape(self):
        s = Storage(tensors=["A"], component="MainMem")
        self.assertEqual(s._render_node_shape(), "cylinder")

    def test_persistent_default_false(self):
        s = Storage(tensors=["A"], component="MainMem")
        self.assertFalse(s.persistent)


# ============================================================================
# Toll (mapping)
# ============================================================================


class TestMappingToll(unittest.TestCase):
    """Tests for Toll mapping node."""

    def test_basic_creation(self):
        t = Toll(tensors=["A"], component="Network")
        self.assertEqual(t.component, "Network")

    def test_str_representation(self):
        t = Toll(tensors=["A"], component="Network")
        result = str(t)
        self.assertIn("Network", result)
        self.assertIn("passes", result)

    def test_render_node_shape(self):
        t = Toll(tensors=["A"], component="Network")
        self.assertEqual(t._render_node_shape(), "rarrow")


# ============================================================================
# Compute (mapping)
# ============================================================================


class TestMappingCompute(unittest.TestCase):
    """Tests for Compute mapping node."""

    def test_basic_creation(self):
        c = Compute(einsum="Matmul", component="MAC")
        self.assertEqual(c.einsum, "Matmul")
        self.assertEqual(c.component, "MAC")

    def test_str_representation(self):
        c = Compute(einsum="Matmul", component="MAC")
        result = str(c)
        self.assertIn("MAC", result)
        self.assertIn("Matmul", result)

    def test_compact_str(self):
        c = Compute(einsum="Matmul", component="MAC")
        result = c.compact_str()
        self.assertIn("MAC", result)
        self.assertIn("Matmul", result)

    def test_render_node_shape(self):
        c = Compute(einsum="Matmul", component="MAC")
        self.assertEqual(c._render_node_shape(), "ellipse")


# ============================================================================
# Reservation
# ============================================================================


class TestReservation(unittest.TestCase):
    """Tests for Reservation mapping node."""

    def test_basic_creation(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        self.assertEqual(r.purposes, ["A"])
        self.assertEqual(r.resource, "MainMem")

    def test_purpose_single(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        self.assertEqual(r.purpose, "A")

    def test_purpose_multiple_raises(self):
        r = Reservation(purposes=["A", "B"], resource="MainMem")
        with self.assertRaises(ValueError):
            _ = r.purpose

    def test_str_representation(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        result = str(r)
        self.assertIn("MainMem", result)
        self.assertIn("A", result)

    def test_compact_str(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        result = r.compact_str()
        self.assertIn("A", result)
        self.assertIn("MainMem", result)

    def test_equality(self):
        r1 = Reservation(purposes=["A"], resource="MainMem")
        r2 = Reservation(purposes=["A"], resource="MainMem")
        self.assertEqual(r1, r2)

    def test_inequality(self):
        r1 = Reservation(purposes=["A"], resource="MainMem")
        r2 = Reservation(purposes=["B"], resource="MainMem")
        self.assertNotEqual(r1, r2)

    def test_persistent_default(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        self.assertFalse(r.persistent)

    def test_render_node_shape(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        self.assertEqual(r._render_node_shape(), "component")


# ============================================================================
# Nested
# ============================================================================


class TestNested(unittest.TestCase):
    """Tests for Nested mapping node."""

    def test_basic_creation(self):
        n = Nested(nodes=[])
        self.assertEqual(len(n.nodes), 0)

    def test_with_children(self):
        n = Nested(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Storage(tensors=["A"], component="MainMem"),
                Compute(einsum="Matmul", component="MAC"),
            ]
        )
        self.assertEqual(len(n.nodes), 3)

    def test_get_nodes_of_type(self):
        n = Nested(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Temporal(rank_variable="n", tile_shape=2),
                Storage(tensors=["A"], component="MainMem"),
                Compute(einsum="Matmul", component="MAC"),
            ]
        )
        temporals = n.get_nodes_of_type(Temporal)
        self.assertEqual(len(temporals), 2)

    def test_flatten(self):
        n = Nested(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Storage(tensors=["A"], component="MainMem"),
            ]
        )
        flat = n._flatten()
        self.assertGreaterEqual(len(flat), 3)  # Nested + 2 children

    def test_compact_str(self):
        n = Nested(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Storage(tensors=["A"], component="MainMem"),
            ]
        )
        result = n.compact_str()
        self.assertIn("m", result)


# ============================================================================
# Split, Sequential, Pipeline
# ============================================================================


class TestSplit(unittest.TestCase):
    """Tests for Split node."""

    def test_basic_creation(self):
        s = Split(nodes=[])
        self.assertEqual(len(s.nodes), 0)

    def test_str_representation(self):
        s = Split(nodes=[])
        self.assertEqual(str(s), "Split")

    def test_render_node_shape(self):
        s = Split(nodes=[])
        self.assertEqual(s._render_node_shape(), "hexagon")


class TestSequential(unittest.TestCase):
    """Tests for Sequential split node."""

    def test_basic_creation(self):
        s = Sequential(nodes=[])
        self.assertEqual(len(s.nodes), 0)

    def test_with_nested_children(self):
        s = Sequential(
            nodes=[
                Nested(nodes=[Compute(einsum="E1", component="MAC")]),
                Nested(nodes=[Compute(einsum="E2", component="MAC")]),
            ]
        )
        self.assertEqual(len(s.nodes), 2)


class TestPipeline(unittest.TestCase):
    """Tests for Pipeline split node."""

    def test_basic_creation(self):
        p = Pipeline(nodes=[])
        self.assertEqual(len(p.nodes), 0)


# ============================================================================
# Mapping (top-level)
# ============================================================================


class TestMapping(unittest.TestCase):
    """Tests for the top-level Mapping class."""

    def test_empty_mapping(self):
        m = Mapping()
        self.assertEqual(len(m.nodes), 0)

    def test_mapping_with_nodes(self):
        m = Mapping(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Storage(tensors=["A"], component="MainMem"),
                Compute(einsum="Matmul", component="MAC"),
            ]
        )
        self.assertEqual(len(m.nodes), 3)

    def test_loops_property(self):
        m = Mapping(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Temporal(rank_variable="n", tile_shape=2),
                Compute(einsum="Matmul", component="MAC"),
            ]
        )
        self.assertEqual(len(m.loops), 2)

    def test_remove_reservations(self):
        m = Mapping(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Reservation(purposes=["A"], resource="MainMem"),
                Compute(einsum="Matmul", component="MAC"),
            ]
        )
        m.remove_reservations()
        self.assertEqual(len(m.nodes), 2)
        self.assertFalse(any(isinstance(n, Reservation) for n in m.nodes))

    def test_split_loop_with_multiple_rank_variables(self):
        m = Mapping(
            nodes=[
                Temporal(rank_variable={"m", "n"}, tile_shape=4),
                Compute(einsum="E", component="MAC"),
            ]
        )
        m.split_loop_with_multiple_rank_variables()
        temporals = [n for n in m.nodes if isinstance(n, Temporal)]
        self.assertEqual(len(temporals), 2)

    def test_split_tensor_holders_with_multiple_tensors(self):
        m = Mapping(
            nodes=[
                Storage(tensors=["A", "B"], component="MainMem"),
                Compute(einsum="E", component="MAC"),
            ]
        )
        m.split_tensor_holders_with_multiple_tensors()
        storages = [n for n in m.nodes if isinstance(n, Storage)]
        self.assertEqual(len(storages), 2)

    def test_get_nodes_of_type(self):
        m = Mapping(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Storage(tensors=["A"], component="MainMem"),
                Compute(einsum="Matmul", component="MAC"),
            ]
        )
        storages = m.get_nodes_of_type(Storage)
        self.assertEqual(len(storages), 1)

    def test_mapping_node_hash(self):
        """Mapping nodes use id-based hashing."""
        n1 = Temporal(rank_variable="m", tile_shape=4)
        n2 = Temporal(rank_variable="m", tile_shape=4)
        # Different objects should have different hashes (identity-based)
        self.assertNotEqual(hash(n1), hash(n2))

    def test_mapping_node_identity_equality(self):
        """Mapping nodes use identity-based equality."""
        n1 = Temporal(rank_variable="m", tile_shape=4)
        n2 = Temporal(rank_variable="m", tile_shape=4)
        # But Loop.__eq__ compares by value, so they're equal by value
        self.assertEqual(n1, n2)


# ============================================================================
# MappingNodeWithChildren utilities
# ============================================================================


class TestMappingNodeWithChildrenUtils(unittest.TestCase):
    """Tests for utility methods on MappingNodeWithChildren."""

    def test_clear_nodes_of_type(self):
        n = Nested(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                Reservation(purposes=["A"], resource="MainMem"),
                Storage(tensors=["A"], component="MainMem"),
                Compute(einsum="E", component="MAC"),
            ]
        )
        n.clear_nodes_of_type(Reservation)
        self.assertFalse(any(isinstance(x, Reservation) for x in n.nodes))
        self.assertEqual(len(n.nodes), 3)

    def test_clear_nodes_specific(self):
        r = Reservation(purposes=["A"], resource="MainMem")
        n = Nested(
            nodes=[
                Temporal(rank_variable="m", tile_shape=4),
                r,
                Compute(einsum="E", component="MAC"),
            ]
        )
        n.clear_nodes(r)
        self.assertEqual(len(n.nodes), 2)


if __name__ == "__main__":
    unittest.main()



import copy
import math
import pandas as pd
from uuid import UUID
from collections import defaultdict

import sympy
from accelforge.frontend.mapping import Loop, Mapping, Spatial, Temporal
from accelforge.frontend.workload import EinsumName
from accelforge.mapper.FFM._join_pmappings.compatibility import (
    Compatibility,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import (
    MAPPING_COLUMN,
    PmappingDataframe,
    col2nameloop,
    col_used_in_pareto,
    is_reservation_col,
    makepareto,
    tensor2col,
    col2nameloop,
    is_reservation_col,
    nameloop2col,
)

from accelforge.frontend.mapper.metrics import Metrics
from accelforge.mapper.FFM._make_pmappings.make_pmappings_from_templates.make_tile_shapes import (
    make_tile_shapes,
    IMPERFECT,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingGroup
from accelforge.mapper.FFM._make_pmappings.pmapper_job import (
    Job,
    SameCompatibilityJobs,
)
from accelforge.mapper.FFM._pareto_df.df_convention import (
    is_fused_loop_col,
    is_n_iterations_col,
)
from accelforge.util._mathfuncs import _count_factorizations


def shift_reservations_by_null_loop_indices(
    mappings: pd.DataFrame, null_loop_indices: set[int]
):
    prev = copy.deepcopy(mappings)  # TODO: Is this needed?
    target2newabovename = {}
    dropcols = []
    for c in mappings.columns:
        if not is_reservation_col(c):
            continue
        name, above = col2nameloop(c)
        new_above = above - sum(above > i for i in null_loop_indices)
        target = nameloop2col(name, new_above)
        if target in target2newabovename:
            if above > target2newabovename[target][1]:
                dropcols.append(nameloop2col(*target2newabovename[target]))
                target2newabovename[target] = (name, above)
            else:
                dropcols.append(c)
        else:
            target2newabovename[target] = (name, above)

    mappings.drop(columns=dropcols, inplace=True)
    renames = {}
    for target, (name, above) in target2newabovename.items():
        renames[nameloop2col(name, above)] = target
    mappings.rename(columns=renames, inplace=True)
    if len(mappings.columns) != len(mappings.columns.unique()):
        shift_reservations_by_null_loop_indices(prev, null_loop_indices)
        raise ValueError(f"Duplicate columns: {mappings.columns}")
    assert len(mappings.columns) == len(mappings.columns.unique())
    return mappings


def get_equivalent_pmappings(
    pmapping_group: PmappingGroup, reservation_levels: set[int]
) -> list[PmappingGroup]:
    equivalent_permutations = pmapping_group.compatibility.make_equivalent_permutations(
        reservation_levels
    )
    result = [PmappingGroup(c, None) for c in equivalent_permutations]
    return result


def mapping2fused_loop_cols(mapping: Mapping, einsum_name: EinsumName):
    cols = []
    for loop in [l for l in mapping.nodes if isinstance(l, Loop) and l._fused]:
        if loop.tile_shape is not None:
            cols.append(loop.tile_shape)
        elif loop.tile_pattern is not None:
            cols.append(loop.tile_pattern.tile_shape)
            cols.append(loop.tile_pattern.initial_tile_shape)
        else:
            raise ValueError(f"Can't find tile shape or tile pattern for loop {loop}")
        return [f"{einsum_name}<SEP>{c}" if isinstance(c, str) else c for c in cols]


def get_fused_loop_indices(
    df: pd.DataFrame,
    compatibility: Compatibility,
    einsum_name: EinsumName,
    return_as_int: bool = False,
) -> pd.Series | int:
    result = []

    loops = compatibility.loops
    for i, loop in enumerate(loops):
        col = loop.tile_pattern.calculated_n_iterations
        assert col is not None, f"Loop {loop} has no calculated n_iterations"
        if isinstance(col, str):
            col = df[f"{einsum_name}<SEP>{col}"]
        elif isinstance(col, sympy.Symbol):
            col = df[f"{einsum_name}<SEP>{col.name}"]
        result.append(col != 1)

    if return_as_int:
        n = 0
        for b in result:
            n = n * 2 + b
        return n
    else:
        r2 = []
        for b in result:
            if len(b.unique()) > 1:
                raise ValueError(f"This won't work if there's more than one")
            r2.append(b.iloc[0])
        return r2


def _count_loops(job: Job) -> tuple[list[int], list[int], dict[str, int]]:
    nodes = job.mapping.nodes
    temporal_n_loops = []
    spatial_n_loops = []
    rv_spatial_count = defaultdict(int)
    rv_temporal_count = defaultdict(int)
    cur_n_loops = 0
    spatial_dim = None

    def pop_loop():
        nonlocal cur_n_loops
        if cur_n_loops >= 1:
            if spatial_dim is not None:
                spatial_n_loops.append(cur_n_loops)
            else:
                temporal_n_loops.append(cur_n_loops)
        cur_n_loops = 0

    for node in nodes:
        cur_spatial_dim = None
        if isinstance(node, Spatial):
            cur_spatial_dim = node.name
            rv_spatial_count[node.rank_variable] += 1
        if cur_spatial_dim != spatial_dim:
            pop_loop()
            spatial_dim = cur_spatial_dim
        if isinstance(node, Loop):
            cur_n_loops += 1
            if isinstance(node, Temporal):
                rv_temporal_count[node.rank_variable] += 1
        else:
            pop_loop()
    pop_loop()
    return temporal_n_loops, spatial_n_loops, rv_spatial_count, rv_temporal_count


def multiply_n_pmappings_by_permutations(n_pmappings: int, job: Job) -> int:
    option = job.spec.mapper._count_option_for_mapsapce_size_evaluation
    # if option == "normal":
    #     return n_pmappings

    temporal_n_loops, spatial_n_loops, rv_spatial_count, rv_temporal_count = (
        _count_loops(job)
    )

    rv = {k: v for k, v in job.rank_variable_bounds.items()}

    if "non_helpful_tile_shapes" in option:
        rv_temporal_count = {r: len(temporal_n_loops) for r in rv.keys()}

    if "non_helpful_loops_for_loop_orders" in option:
        for i in range(len(temporal_n_loops)):
            temporal_n_loops[i] = len(rv)

    # Count number of tile shapes
    rv2loops = {r: rv_spatial_count[r] + rv_temporal_count[r] for r in rv}
    n_factorizations = math.prod(
        _count_factorizations(b, rv2loops[r], imperfect=IMPERFECT)
        for r, b in rv.items()
    )
    n_temporal_loop_orders = math.prod(math.factorial(n) for n in temporal_n_loops)

    n = n_factorizations

    # assert n >= n_pmappings, f"n_pmappings: {n_pmappings} > n: {n}"

    if "redundant_loop_orders" in option:
        # job.mapping._n_loop_orders is the number of permutations that we actually
        # evaluate. Don't want to double count them.
        n *= n_temporal_loop_orders / job.mapping._n_loop_orders

    # assert n >= n_pmappings, f"n_pmappings: {n_pmappings} > n: {n}"

    return n


def assert_all_jobs_have_same_symbols(
    jobs_with_similar_compatibilities: SameCompatibilityJobs,
):
    iteration2symbols = []
    for j in jobs_with_similar_compatibilities:
        for t in j.compatibility.tensors:
            for i, l in enumerate(t.loops):
                if len(iteration2symbols) <= i:
                    iteration2symbols.append(set())
                iteration2symbols[i].add(l.tile_pattern.calculated_n_iterations)
    assert all(
        len(s) == 1 for s in iteration2symbols
    ), "All jobs must have the same symbols for compatibility n_iterations"


def make_pmappings_from_templates(
    jobs_with_similar_compatibilities: SameCompatibilityJobs,
) -> tuple[EinsumName, list[PmappingGroup], dict[UUID, Mapping], SameCompatibilityJobs]:
    jwsc = jobs_with_similar_compatibilities

    results = []

    for job in jobs_with_similar_compatibilities:
        try:
            result, tensor2mapping = make_tile_shapes(job)
        except Exception as e:
            e.add_note(f"Einsum {jwsc.einsum_name} compatibility {job.compatibility}")
            raise
        job.compatibility = job.compatibility.populate_loops()

        # Ctrl-F for CONTIGUOUS_ITERATION_SPACE_DISCUSSION TODO: Turn tensor2pmapping
        # into per-tensor compatibility

        # This changes the pmapping count to include superfluous permutations
        # TODO: Add a multiplier for the permutations that we include in the fusion
        # piece, which are NOT known to be superfluous

        # prev = job.spec.mapper._count_option_for_mapsapce_size_evaluation
        # job.spec.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders_and_irrelevant_loops"
        # a = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)
        # job.spec.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders"
        # b = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)

        # if a < b:
        #     job.spec.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders_and_irrelevant_loops"
        #     a = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)
        #     job.spec.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders"
        #     b = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)
        #     assert False

        # job.spec.mapper._count_option_for_mapsapce_size_evaluation = prev
        job.n_total_pmappings = multiply_n_pmappings_by_permutations(
            job.n_total_pmappings, job
        )

        result[MAPPING_COLUMN] = job.job_id
        cols_to_drop = []
        for col in result.columns:
            if is_reservation_col(col):
                resource = col2nameloop(col)[0]
                if resource in job.memories_track_pmappings_only:
                    cols_to_drop.append(col)
        result.drop(columns=cols_to_drop, inplace=True)
        results.append(result)

    fusable_tensors = jwsc.fusable_tensors
    einsum_name = jwsc.einsum_name
    metrics = jwsc.metrics
    limit_capacity_drop_valid_reservations = not (Metrics.RESOURCE_USAGE & metrics)
    compatibility = jwsc.compatibility

    # Creating a PmappingDataframe fills in reservation columns since different pmappings
    # have different ones.
    next_shared_loop_index = compatibility.n_loops - 1
    df = PmappingDataframe.concat(
        [
            PmappingDataframe(
                r,
                skip_pareto=True,
                next_shared_loop_index=next_shared_loop_index,
                n_total_pmappings=1,  # Unused for now, just making an initial Pareto
                n_valid_pmappings=1,  # Unused for now, just making an initial Pareto
                ignored_resources=job.ignored_resources,
                # False because we may have lifetimes that stretch through this Einsum
                # due to data dependencies, not loops
                limit_capacity_drop_valid_reservations=False,
            )
            for r in results
        ],
        skip_pareto=True,
    ).data
    if df.empty:
        return einsum_name, [], {}, jobs_with_similar_compatibilities

    tensor_cols = [tensor2col(tensor) for tensor in fusable_tensors]
    df.columns = [
        c if col_used_in_pareto(c) or c in tensor_cols else f"{einsum_name}<SEP>{c}"
        for c in df.columns
    ]

    fused_loop_cols = [
        f"{einsum_name}<SEP>{c}"
        for c in compatibility.symbols()
        if not is_n_iterations_col(c)
    ]

    job0 = next(iter(jobs_with_similar_compatibilities))

    # Pareto prune
    df = makepareto(df, split_by_cols=fused_loop_cols).copy()

    jobs_passed_pareto = sorted(df[f"{einsum_name}<SEP>{MAPPING_COLUMN}"].unique())
    pmapping_objects = {
        job.job_id: job.mapping
        for job in jobs_with_similar_compatibilities
        if job.job_id in jobs_passed_pareto
    }

    assert_all_jobs_have_same_symbols(jobs_with_similar_compatibilities)
    # Otherwise, following logic fails

    df["fused_loop_indices"] = get_fused_loop_indices(
        df, job0.compatibility, einsum_name, return_as_int=True
    )
    groups = list(df.groupby(["fused_loop_indices"]))
    total_pmappings_per_group = sum(
        j.n_total_pmappings for j in jobs_with_similar_compatibilities
    ) / len(groups)
    valid_pmappings_per_group = sum(
        j.n_valid_pmappings for j in jobs_with_similar_compatibilities
    ) / len(groups)

    pmapping_groups = []
    for _, mappings in groups:
        compatibility = jwsc.compatibility
        fused_loop_indices = []

        for i, f in enumerate(
            get_fused_loop_indices(
                mappings, compatibility, einsum_name, return_as_int=False
            )
        ):
            if f:
                fused_loop_indices.append(i)

        null_loop_indices = tuple(
            i for i in range(compatibility.n_loops) if i not in fused_loop_indices
        )

        dropcols = ["fused_loop_indices"]
        mappings = mappings.drop(columns=dropcols)

        compatibility = compatibility.drop_loop_indices(null_loop_indices)

        symbol_renames, compatibility = compatibility.make_fused_loop_symbols(
            einsum_name
        )
        for k, v in symbol_renames.items():
            mappings[v] = mappings[f"{einsum_name}<SEP>{k}"]
        shift_reservations_by_null_loop_indices(mappings, null_loop_indices)

        symbols = compatibility.symbols()
        dropcols = [
            c for c in mappings.columns if is_fused_loop_col(c) and c not in symbols
        ]
        mappings = mappings.drop(columns=dropcols)

        energy_cols = [c for c in mappings.columns if "Total<SEP>energy" in c]
        if (mappings[energy_cols] < 0).any(axis=None):
            mapping_with_negative_energy = mappings[
                (mappings[energy_cols] < 0).any(axis=1)
            ]
            msg = ""
            for _, row in mapping_with_negative_energy.iterrows():
                for k, v in row.items():
                    msg += f"{k}: {v}\n"
                msg += "\n"
            raise RuntimeError(f"negative energy:\n{msg}")

        # Skip pareto because we already did it above
        next_shared_loop_index_this_group = compatibility.n_loops - 1
        partial_mappings = PmappingDataframe(
            mappings,
            next_shared_loop_index=next_shared_loop_index_this_group,
            n_total_pmappings=total_pmappings_per_group,
            n_valid_pmappings=valid_pmappings_per_group,
            skip_pareto=next_shared_loop_index_this_group == next_shared_loop_index,
            ignored_resources=job.ignored_resources,
            # False because we may have lifetimes that stretch through this Einsum
            # due to data dependencies, not loops
            limit_capacity_drop_valid_reservations=False,
        )
        pmapping_groups.append(PmappingGroup(compatibility, partial_mappings))

    return (
        einsum_name,
        pmapping_groups,
        pmapping_objects,
        jobs_with_similar_compatibilities,
    )



import copy
import functools
import logging
import os
import glob
import re
import io
from typing import Callable, List, Dict, Any, OrderedDict, Tuple
import ruamel.yaml
import warnings
from ruamel.yaml.error import ReusedAnchorWarning
from jinja2 import StrictUndefined, Environment, FileSystemLoader
import threading
import time


PARSING_LOCK = threading.Lock()
THREAD_ID = 0

SCRIPTS_FROM = []
EXTRA_PLUG_IN_PATHS = []


class LockAcquirer:
    def __init__(self):
        self.has_lock = False

    def __enter__(self):
        while True:
            if PARSING_LOCK.acquire(blocking=False):
                global THREAD_ID
                THREAD_ID = threading.get_ident()
                self.has_lock = True
                break
            if THREAD_ID == threading.get_ident():
                break
            time.sleep(0.01)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.has_lock:
            PARSING_LOCK.release()


def recursive_mutator_stop(func):
    return func
    cache = set()

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        assert not kwargs, (
            f"Recursive mutator stop only works with non-keyword arguments. "
            f"Args were {args} and kwargs were {kwargs}."
        )
        k = id(args[0])
        if k in cache:
            return args[0]
        cache.add(k)
        try:
            result = func(*args, **kwargs)
        finally:
            cache.remove(k)
        return result

    return wrapper


def recursive_mutator_eq_stop(func):
    return func
    cache = set()

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if args[0] in cache:
            failstr = (
                f"Infinite recursion detected: {func.__name__}("
                f"{args[0]}) called from within itself. Does a YAML file "
                f"include itself?"
            )
            raise RuntimeError(failstr)
        cache.add(args[0])
        try:
            result = func(*args, **kwargs)
        finally:
            cache.remove(args[0])
        return result

    return wrapper


class MultiIncludeWrapper:
    def __init__(self, contents: List):
        self.contents = contents


def append_path(p: str, cur_path: str, include_dirs: List[str]):
    new_paths = find_paths(p, cur_path, include_dirs)
    include_dirs += new_paths
    logging.info(f"YAML Adding {new_paths} to include paths")
    return ""


def find_paths(p: str, cur_path: str, include_dirs: List[str]):
    if isinstance(p, list):
        paths = [find_paths(x, cur_path, include_dirs) for x in p]
    else:
        searched = []
        paths = []
        prepend = (
            [""] if os.path.isabs(p) else [os.path.dirname(cur_path)] + include_dirs
        )

        for d in prepend:
            s = os.path.abspath(os.path.realpath(os.path.join(d, p)))
            globbed_paths = glob.glob(s)
            if globbed_paths:
                paths.extend(globbed_paths)
            searched.append(s)
        if not paths:
            raise FileNotFoundError(
                f"Could not find file {p} in any of the following paths:"
                + "\n  "
                + "\n  ".join(searched)
            )

    unique_paths = []
    uniques = set()
    while paths:
        p = os.path.realpath(os.path.abspath(paths.pop(0)))
        if p not in uniques:
            unique_paths.append(p)
            uniques.add(p)
    return unique_paths


def find_path(p: str, cur_path: str, include_dirs: List[str]):
    prepend = [""] if os.path.isabs(p) else [os.path.dirname(cur_path)] + include_dirs
    searched = []
    for d in prepend:
        s = os.path.abspath(os.path.realpath(os.path.join(d, p)))
        if os.path.exists(s):
            return s
        searched.append(s)
    raise FileNotFoundError(
        f"Could not find file {p} in any of the following paths:"
        + "\n  "
        + "\n  ".join(searched)
    )


@recursive_mutator_eq_stop
def load_file_and_includes(
    path: str,
    data: Dict[str, Any] = None,
    include_dirs: List[str] = None,
) -> Tuple[str, Dict[str, Any]]:
    """
    Load a YAML file and recursively load any included YAML files
    :param path: string that specifies the path of the YAML file to be loaded
    :param data: dictionary that contains the data to be rendered
    :param include_dirs: list of directories to search for included files
    :return: string that contains the loaded YAML content
    """
    path = os.path.abspath(os.path.realpath(path))
    if not os.path.exists(path):
        raise FileNotFoundError(f"Could not find file {path}")

    data = data or {}
    data = {k: v for k, v in data.items()}
    include_dirs = include_dirs or []
    include_dirs = [d for d in include_dirs]

    include_counter = 0

    def include(p, single, indices: str = ""):
        # If the path is a relative path, make it relative to the current file
        to_include = []
        indices = indices.lstrip(".")
        nonlocal include_counter
        include_name = (
            os.path.basename(path).rsplit(".", 1)[0] + "_" + str(include_counter)
        )
        include_name = re.sub(r"\W+", "", include_name)
        for np in find_paths(p, path, include_dirs):
            logging.info(
                f"YAML Adding {np} to document with !include{'_all' if not single else ''}"
            )
            to_include.append(load_yaml(np, data, include_dirs))

        if single:
            if len(to_include) > 1:
                raise RuntimeError(
                    f"More than one file found for {path}: {find_paths(path)}."
                    f"To include multiple files, use include_all()."
                )

        data[include_name] = (
            to_include[0] if len(to_include) == 1 else MultiIncludeWrapper(to_include)
        )
        v = f"!include_loaded {include_name}"
        if indices:
            v += "." + indices
        include_counter += 1
        return v

    def include_single(p, indices: str = ""):
        return include(p, True, indices)

    def include_all(p, indices: str = ""):
        return include(p, False, indices)

    def include_text(p):
        found = []
        for np in find_paths(p, path, include_dirs):
            found.append(load_file_and_includes(np, data, include_dirs)[0])
            logging.info(f"YAML Adding {np} to document with !include_text")
        return "\n".join(found)

    # Add include_as to the template environment
    env = Environment(
        loader=FileSystemLoader(os.path.dirname(path)), undefined=StrictUndefined
    )

    def setenv(key, value):
        key, value = str(key), str(value)
        os.environ[key] = value
        return "{{ setenv('" + key + "', '" + value + "') }}}}"

    def path_exists(p):
        try:
            find_path(p, path, include_dirs)
            return True
        except FileNotFoundError:
            return False

    env.globals["cwd"] = lambda: os.path.dirname(path)
    env.globals["include"] = include_single
    env.globals["include_all"] = include_all
    env.globals["include_text"] = include_text
    env.globals["find_path"] = lambda x: find_path(x, path, include_dirs)
    env.globals["find_paths"] = lambda x: find_paths(x, path, include_dirs)
    env.globals["path_exists"] = path_exists
    env.globals["setenv"] = setenv

    env.globals["add_to_path"] = lambda p: append_path(p, path, include_dirs)

    path_file = path[len(os.path.dirname(path)) + 1 :]
    template = env.get_template(path_file)
    string = template.render(data, undefined=StrictUndefined)
    return string, data, include_dirs


@recursive_mutator_stop
def merge_check(x: dict[str, Any] | list[Any] | Any) -> None:
    if isinstance(x, list):
        for i, v in enumerate(x):
            x[i] = merge_check(v)
    elif isinstance(x, dict):
        found_merge = False
        for k, v in list(x.items()):
            x[k] = merge_check(v)
            if str(k) == "<<<" or str(k) == "<<":
                assert not found_merge, (
                    f'Cannot have multiple "<<<" or "<<" keys in a dict. '
                    f"Keys were {list(x.keys())}"
                )
                found_merge = True
                x = merge(x, copy.deepcopy(x.pop(k)), str(k) == "<<<")
    return x


ERRCOUNT = 0


def load_yaml(
    path: str,
    data: Dict[str, Any] = None,
    include_dirs: List[str] = None,
) -> Dict[str, Any]:
    """
    Load YAML content from a file or string
    :param path: string that specifies the path of the YAML file to be loaded
    :param data: dictionary that contains the data to be rendered
    :param include_dirs: list of directories to search for included files
    :return: evaluated YAML content or YAML object
    """
    with LockAcquirer():
        data = {k: v for k, v in data.items()} if data is not None else {}
        path = os.path.abspath(os.path.realpath(path))
        try:
            evaluated, data, include_dirs = load_file_and_includes(
                path, data, include_dirs
            )
        except Exception as e:
            e.add_note(f"Error loading YAML file {path}")
            raise
        try:
            result = merge_check(get_yaml(path, data).load(evaluated))
            return result
        except Exception as e:
            global ERRCOUNT
            failpath = f"/tmp/yamleval_error{ERRCOUNT}.yaml"
            ERRCOUNT += 1
            with open(failpath, "w") as f:
                f.write(evaluated)
            e.add_note(
                f"Error parsing YAML file {path}. Offending file written to "
                f"{failpath}"
            )
            raise


@recursive_mutator_stop
def merge(
    merge_into: dict, tomerge: dict | list | tuple, recursive: bool = True
) -> dict:
    if isinstance(tomerge, (list, tuple)):
        combined = dict()
        for m in tomerge:
            combined = merge(combined, m, recursive)
        tomerge = combined
    if not isinstance(tomerge, dict):
        raise ValueError(
            f'Expected a dict under the "<<<" or "<<" keys, but ' f"got {tomerge}"
        )
    if not isinstance(merge_into, dict):
        raise ValueError(
            f'Expected to merge into a dict with the "<<<" key, '
            f"but got {merge_into}"
        )

    for k, v in tomerge.items():
        if k not in merge_into:
            merge_into[k] = v
        elif not recursive:
            continue
        elif isinstance(merge_into[k], (NoMergeListWrapper, NoMergeDictWrapper)):
            continue
        elif isinstance(merge_into[k], dict) and isinstance(v, dict):
            merge_into[k] = merge(merge_into[k], v, recursive)
        elif isinstance(merge_into[k], list) and isinstance(v, list):
            merge_into[k] = merge_into[k] + v
    return merge_into


def represent_none(self, data: None) -> str:
    """
    Represent None as 'null' in YAML
    :param self: YAML representer object
    :param data: None object to be represented
    :return: 'null' string
    """
    return self.represent_scalar("tag:yaml.org,2002:null", "null")


def ordereddict_to_dict(self, dictionary: OrderedDict) -> Dict[str, Any]:
    """
    Change an OrderedDict to a dictionary in YAML
    :param self: YAML representer object
    :param dictionary: OrderedDict object to be represented
    :return: dictionary object
    """
    d = {}
    for key in dictionary.keys():
        d[key] = dictionary[key]
    return self.represent_dict(dictionary)


@recursive_mutator_stop
def recursive_unorder_dict(to_unorder: Dict[str, Any]) -> Dict[str, Any]:
    if isinstance(to_unorder, dict):
        return {k: recursive_unorder_dict(v) for k, v in to_unorder.items()}
    elif isinstance(to_unorder, list):
        return [recursive_unorder_dict(v) for v in to_unorder]
    return to_unorder


@recursive_mutator_stop
def callables2strings(to_convert: Dict[str, Any]) -> Dict[str, Any]:
    if isinstance(to_convert, dict):
        to_convert = {k: callables2strings(v) for k, v in to_convert.items()}
    elif isinstance(to_convert, list):
        to_convert = [callables2strings(v) for v in to_convert]
    elif isinstance(to_convert, Callable):
        to_convert = str(getattr(to_convert, "_original_expression", to_convert))
    return to_convert


def write_yaml_file(filepath: str, content: Dict[str, Any]) -> None:
    """
    Write YAML content to a file
    :param filepath: string that specifies the destination file path
    :param content: YAML string that needs to be written to the destination file
    :return: None
    """
    if os.path.exists(filepath):
        os.remove(filepath)
    if os.path.dirname(filepath):
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
    out_file = open(filepath, "a")
    out_file.write(to_yaml_string(content))


def to_yaml_string(content: Dict[str, Any]) -> str:
    """
    Convert YAML content to a string
    :param content: YAML content to be converted to a string
    :return: string representation of the YAML content
    """
    with LockAcquirer():
        dumpstream = io.StringIO()
        get_base_yaml().dump(
            callables2strings(recursive_unorder_dict(content)), stream=dumpstream
        )
        return dumpstream.getvalue()


def get_base_yaml() -> ruamel.yaml.YAML:
    yaml = ruamel.yaml.YAML(typ="rt")
    # yaml.default_flow_style = None
    yaml.indent(mapping=4, sequence=4, offset=2)
    yaml.preserve_quotes = True

    def recursive_mutator_stop(func):
        cache = set()

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            assert not kwargs, (
                f"Recursive mutator stop only works with non-keyword "
                f"arguments. Args were {args} and kwargs were {kwargs}."
            )
            k = id(args[0])
            if k in cache:
                return args[0]
            cache.add(k)
            try:
                result = func(*args, **kwargs)
            finally:
                cache.remove(k)
            return result

        return wrapper

    yaml.representer.add_representer(type(None), recursive_mutator_stop(represent_none))
    yaml.representer.add_representer(
        OrderedDict, recursive_mutator_stop(ordereddict_to_dict)
    )

    return yaml


def get_yaml(path: str, data: Dict[str, Any] = None) -> ruamel.yaml.YAML:
    """Get a YAML object with the right settings"""
    yaml = get_base_yaml()
    ymf = YAMLFileLoader(path, data)
    # yaml.default_flow_style = None

    warnings.simplefilter("ignore", ReusedAnchorWarning)
    yaml.constructor.add_constructor("!include_loaded", ymf.include_loaded)
    yaml.constructor.add_constructor("!include", ymf.include)
    yaml.constructor.add_constructor("!includedir", ymf.includedir)
    yaml.constructor.add_constructor("!nomerge", ymf.nomerge)

    return yaml


class NoMergeListWrapper(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NoMergeDictWrapper(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class YAMLFileLoader:
    def __init__(self, path: str, data: Dict[str, Any] = None) -> None:
        self.path = path
        self.data = data or {}
        self.include_counter = 0
        self.loading_from_dir = os.path.abspath(os.path.dirname(path))
        self.include_data = data or {}
        self.env = Environment(
            loader=FileSystemLoader(os.path.dirname(path)), undefined=StrictUndefined
        )

    def nomerge(self, constructor, node):
        # print(f"Got node {node}")
        # Pop the tag
        # node.tag = None
        # print(f"Got node {node}")
        # # value = constructor.construct_object(node, deep=True)
        # print(f"COnstructed object {value}")
        if isinstance(node, ruamel.yaml.nodes.SequenceNode):
            return NoMergeListWrapper(constructor.construct_sequence(node, deep=True))
        if isinstance(node, ruamel.yaml.nodes.MappingNode):
            return NoMergeDictWrapper(
                ruamel.yaml.SafeConstructor.construct_mapping(
                    constructor, node, deep=True
                )
            )
        raise ValueError(f"!nomerge tag must be applied to a list or dict, not {node}")

    def include_loaded(
        self,
        constructor: ruamel.yaml.constructor.Constructor,
        node: ruamel.yaml.nodes.ScalarNode,
    ) -> dict[str, Any] | None:
        """
        Constructor that parses the !include_loaded relative_file_path and loads the file
        from relative_file_path
        :param self: YAML constructor object
        :param node: YAML node object
        :return: evaluated YAML content
        """
        x = constructor.construct_scalar(node)
        found = self.include_data
        for k in str(x).split("."):
            try:
                if isinstance(found, MultiIncludeWrapper):
                    for i, f in enumerate(found.contents):
                        found.contents[i] = f[k]
                else:
                    found = found[k]
            except (KeyError, TypeError) as e:
                if isinstance(found, MultiIncludeWrapper):
                    e.add_note(
                        f"Could not parse !include_loaded {x}: {k} not found " f"in {f}"
                    )
                    raise
                e.add_note(
                    f"Could not parse !include_loaded {x}: {k} not found "
                    f"in {list(found.keys())}"
                )
                raise
        return found.contents if isinstance(found, MultiIncludeWrapper) else found

    def include(
        self,
        constructor: ruamel.yaml.constructor.Constructor,
        node: ruamel.yaml.nodes.ScalarNode,
    ) -> dict[str, Any] | None:
        """
        Constructor that parses !include relative_file_path and loads the file
        from relative_file_path
        :param self: YAML constructor object
        :param node: YAML node object
        :return: evaluated YAML content
        """
        filepath = constructor.construct_scalar(node)
        if filepath[-1] == ",":
            filepath = filepath[:-1]
        load_from = self.loading_from_dir
        return load_yaml(os.path.join(load_from, filepath), self.include_data)

    def includedir(
        self,
        constructor: ruamel.yaml.constructor.Constructor,
        node: ruamel.yaml.nodes.ScalarNode,
    ) -> list[dict[str, Any]]:
        """
        Constructor that parses the !includedir relative_file_path and loads the
        file from relative_file_path
        :param self: YAML constructor object
        :param node: YAML node object
        :return: list of evaluated YAML contents
        """
        filepath = constructor.construct_scalar(node)
        if filepath[-1] == ",":
            filepath = filepath[:-1]
        dirname = os.path.join(self.loading_from_dir, filepath)
        yamllist = []
        for filename in glob.glob(dirname + "/*.yaml"):
            yamllist.append(load_yaml(filename, self.include_data))
        return yamllist


yaml = get_base_yaml()



"""
Tests for rename frontend classes:
  - Rename
  - RenameList
  - rename_list_factory
  - EinsumRename
  - Renames
"""

import unittest

from accelforge.frontend.renames import (
    Rename,
    RenameList,
    rename_list_factory,
    EinsumRename,
    Renames,
)


# ============================================================================
# Rename
# ============================================================================


class TestRename(unittest.TestCase):
    """Tests for the Rename model."""

    def test_basic_creation(self):
        r = Rename(name="weight", source="W")
        self.assertEqual(r.name, "weight")
        self.assertEqual(r.source, "W")

    def test_expected_count_none_default(self):
        r = Rename(name="input", source="I")
        self.assertIsNone(r.expected_count)

    def test_expected_count_set(self):
        r = Rename(name="output", source="Outputs", expected_count=1)
        self.assertEqual(r.expected_count, 1)

    def test_set_expression_source(self):
        r = Rename(name="input", source="Inputs & Intermediates")
        self.assertEqual(r.source, "Inputs & Intermediates")

    def test_complement_set_expression(self):
        r = Rename(name="weight", source="~(input | output)")
        self.assertEqual(r.source, "~(input | output)")


# ============================================================================
# rename_list_factory
# ============================================================================


class TestRenameListFactory(unittest.TestCase):
    """Tests for rename_list_factory."""

    def test_dict_input(self):
        result = rename_list_factory({"input": "I", "output": "O"})
        self.assertIsInstance(result, RenameList)
        self.assertEqual(len(result), 2)
        names = {r.name for r in result}
        self.assertEqual(names, {"input", "output"})

    def test_list_input(self):
        renames = [
            Rename(name="input", source="I"),
            Rename(name="output", source="O"),
        ]
        result = rename_list_factory(renames)
        self.assertIsInstance(result, RenameList)
        self.assertEqual(len(result), 2)

    def test_empty_dict(self):
        result = rename_list_factory({})
        self.assertIsInstance(result, RenameList)
        self.assertEqual(len(result), 0)

    def test_empty_list(self):
        result = rename_list_factory([])
        self.assertIsInstance(result, RenameList)
        self.assertEqual(len(result), 0)

    def test_invalid_type_raises(self):
        with self.assertRaises(TypeError):
            rename_list_factory(123)

    def test_invalid_type_string_raises(self):
        with self.assertRaises(TypeError):
            rename_list_factory("not_valid")

    def test_dict_preserves_sources(self):
        result = rename_list_factory({"weight": "~(input | output)"})
        self.assertEqual(result[0].source, "~(input | output)")
        self.assertIsNone(result[0].expected_count)


# ============================================================================
# RenameList
# ============================================================================


class TestRenameList(unittest.TestCase):
    """Tests for RenameList."""

    def test_creation(self):
        rl = RenameList(
            [
                Rename(name="a", source="A"),
                Rename(name="b", source="B"),
            ]
        )
        self.assertEqual(len(rl), 2)

    def test_dict_conversion(self):
        rl = RenameList(
            [
                Rename(name="a", source="A"),
                Rename(name="b", source="B"),
            ]
        )
        d = rl.__dict__()
        self.assertEqual(d, {"a": "A", "b": "B"})

    def test_empty_rename_list(self):
        rl = RenameList()
        self.assertEqual(len(rl), 0)

    def test_contains_by_name(self):
        rl = RenameList(
            [
                Rename(name="input", source="I"),
            ]
        )
        # The __contains__ for EvalableList uses name
        self.assertIn("input", rl)
        self.assertNotIn("output", rl)


# ============================================================================
# EinsumRename
# ============================================================================


class TestEinsumRename(unittest.TestCase):
    """Tests for EinsumRename."""

    def test_basic_creation(self):
        er = EinsumRename(name="Matmul")
        self.assertEqual(er.name, "Matmul")
        self.assertEqual(len(er.tensor_accesses), 0)
        self.assertEqual(len(er.rank_variables), 0)

    def test_with_tensor_accesses_dict(self):
        er = EinsumRename(
            name="Matmul",
            tensor_accesses={"input": "A", "output": "C"},
        )
        self.assertEqual(len(er.tensor_accesses), 2)

    def test_with_tensor_accesses_list(self):
        er = EinsumRename(
            name="Matmul",
            tensor_accesses=[
                Rename(name="input", source="A"),
                Rename(name="output", source="C"),
            ],
        )
        self.assertEqual(len(er.tensor_accesses), 2)

    def test_with_rank_variables_dict(self):
        er = EinsumRename(
            name="Matmul",
            rank_variables={"reduced": "k"},
        )
        self.assertEqual(len(er.rank_variables), 1)

    def test_default_name(self):
        er = EinsumRename(name="default")
        self.assertEqual(er.name, "default")


# ============================================================================
# Renames
# ============================================================================


class TestRenames(unittest.TestCase):
    """Tests for the top-level Renames model."""

    def test_empty_renames(self):
        r = Renames()
        self.assertEqual(len(r.einsums), 0)

    def test_with_default_einsum_rename(self):
        r = Renames(
            einsums=[
                EinsumRename(
                    name="default",
                    tensor_accesses={"input": "Inputs & Intermediates"},
                ),
            ]
        )
        self.assertEqual(len(r.einsums), 1)

    def test_get_renames_for_einsum_default_applied(self):
        r = Renames(
            einsums=[
                EinsumRename(
                    name="default",
                    tensor_accesses=[
                        Rename(name="input", source="Inputs & Intermediates"),
                    ],
                ),
            ]
        )
        result = r.get_renames_for_einsum("SomeEinsum")
        self.assertEqual(result.name, "SomeEinsum")
        # The default tensor rename should be applied
        self.assertIn("input", result.tensor_accesses)

    def test_get_renames_for_einsum_without_default(self):
        r = Renames()
        result = r.get_renames_for_einsum("SomeEinsum")
        self.assertEqual(result.name, "SomeEinsum")
        self.assertEqual(len(result.tensor_accesses), 0)

    def test_non_default_einsum_renames_applied_at_eval_time(self):
        """Non-default einsum renames are only resolved during full spec
        evaluation (name-based lookup requires EvalableList). Pre-evaluation,
        get_renames_for_einsum only applies defaults."""
        r = Renames(
            einsums=[
                EinsumRename(
                    name="Matmul",
                    tensor_accesses=[
                        Rename(name="weight", source="W"),
                    ],
                ),
            ]
        )
        # Without evaluation, 'Matmul' is not found in the plain list,
        # so a fresh EinsumRename is created with no tensor_accesses.
        result = r.get_renames_for_einsum("Matmul")
        self.assertEqual(result.name, "Matmul")

    def test_default_applied_when_no_specific_match(self):
        """When a specific einsum is not found, defaults are still applied."""
        r = Renames(
            einsums=[
                EinsumRename(
                    name="default",
                    tensor_accesses=[
                        Rename(name="input", source="DefaultInput"),
                    ],
                ),
            ]
        )
        result = r.get_renames_for_einsum("SomeOtherEinsum")
        input_rename = [rn for rn in result.tensor_accesses if rn.name == "input"]
        self.assertEqual(len(input_rename), 1)
        self.assertEqual(input_rename[0].source, "DefaultInput")


if __name__ == "__main__":
    unittest.main()



"""
Tests for architecture flattening:
  - Arch._flatten (Hierarchical and Fork)
  - Spec._get_flattened_architecture
  - get_nodes_of_type
  - find
  - Duplicate name detection
  - Fork branch selection
"""

import unittest
from pathlib import Path

from accelforge.frontend.spec import Spec
from accelforge.frontend.arch import (
    Arch,
    Hierarchical,
    Fork,
    Memory,
    Compute,
    Fanout,
    Toll,
    Leaf,
    Branch,
)
from accelforge.frontend.workload import Workload

_REPO_ROOT = Path(__file__).parent.parent.parent
EXAMPLES_DIR = _REPO_ROOT / "examples"


def _simple_arch():
    """Simple 3-level architecture: MainMemory -> GlobalBuffer -> MAC."""
    return Arch(
        nodes=[
            Memory(
                name="MainMemory",
                size=1_000_000,
                actions=[
                    {"name": "read", "energy": 1, "latency": 0},
                    {"name": "write", "energy": 1, "latency": 0},
                ],
                leak_power=0,
                area=0,
            ),
            Memory(
                name="GlobalBuffer",
                size=100_000,
                actions=[
                    {"name": "read", "energy": 0.5, "latency": 0},
                    {"name": "write", "energy": 0.5, "latency": 0},
                ],
                leak_power=0,
                area=0,
            ),
            Compute(
                name="MAC",
                actions=[{"name": "compute", "energy": 1, "latency": 1}],
                leak_power=0,
                area=0,
            ),
        ]
    )


def _simple_workload():
    """Single matmul workload."""
    return Workload(
        rank_sizes={"M": 64, "K": 32, "N": 64},
        bits_per_value={"All": 8},
        einsums=[
            {
                "name": "Matmul",
                "tensor_accesses": [
                    {"name": "A", "projection": ["m", "k"]},
                    {"name": "B", "projection": ["k", "n"]},
                    {"name": "C", "projection": ["m", "n"], "output": True},
                ],
            }
        ],
    )


def _fork_arch():
    """Arch with a fork: MainMemory -> GlobalBuffer -> Fork([ScalarUnit], [Fanout -> Register -> MAC])."""
    return Arch(
        nodes=[
            Memory(
                name="MainMemory",
                size=1_000_000,
                actions=[
                    {"name": "read", "energy": 1, "latency": 0},
                    {"name": "write", "energy": 1, "latency": 0},
                ],
                leak_power=0,
                area=0,
            ),
            Memory(
                name="GlobalBuffer",
                size=100_000,
                actions=[
                    {"name": "read", "energy": 0.5, "latency": 0},
                    {"name": "write", "energy": 0.5, "latency": 0},
                ],
                leak_power=0,
                area=0,
            ),
            Fork(
                nodes=[
                    Compute(
                        name="ScalarUnit",
                        actions=[{"name": "compute", "energy": 0, "latency": 1}],
                        leak_power=0,
                        area=0,
                    ),
                ]
            ),
            Fanout(
                name="ArrayFanout",
                spatial=[
                    {
                        "name": "reuse_input",
                        "fanout": 4,
                    },
                ],
            ),
            Memory(
                name="Register",
                size=8,
                actions=[
                    {"name": "read", "energy": 0, "latency": 0},
                    {"name": "write", "energy": 0, "latency": 0},
                ],
                leak_power=0,
                area=0,
            ),
            Compute(
                name="MAC",
                actions=[{"name": "compute", "energy": 1, "latency": 1}],
                leak_power=0,
                area=0,
            ),
        ]
    )


# ============================================================================
# Arch.find
# ============================================================================


class TestArchFind(unittest.TestCase):
    def test_find_memory(self):
        arch = _simple_arch()
        node = arch.find("MainMemory")
        self.assertIsInstance(node, Memory)
        self.assertEqual(node.name, "MainMemory")

    def test_find_compute(self):
        arch = _simple_arch()
        node = arch.find("MAC")
        self.assertIsInstance(node, Compute)
        self.assertEqual(node.name, "MAC")

    def test_find_nonexistent_raises(self):
        arch = _simple_arch()
        with self.assertRaises(ValueError):
            arch.find("NonexistentNode")

    def test_find_in_fork(self):
        arch = _fork_arch()
        node = arch.find("ScalarUnit")
        self.assertIsInstance(node, Compute)
        self.assertEqual(node.name, "ScalarUnit")

    def test_find_after_fork(self):
        arch = _fork_arch()
        node = arch.find("Register")
        self.assertIsInstance(node, Memory)

    def test_find_deep_compute(self):
        arch = _fork_arch()
        node = arch.find("MAC")
        self.assertIsInstance(node, Compute)


# ============================================================================
# Arch.get_nodes_of_type
# ============================================================================


class TestGetNodesOfType(unittest.TestCase):
    def test_get_all_memory(self):
        arch = _simple_arch()
        memories = list(arch.get_nodes_of_type(Memory))
        self.assertEqual(len(memories), 2)
        names = {m.name for m in memories}
        self.assertEqual(names, {"MainMemory", "GlobalBuffer"})

    def test_get_all_compute(self):
        arch = _simple_arch()
        computes = list(arch.get_nodes_of_type(Compute))
        self.assertEqual(len(computes), 1)
        self.assertEqual(computes[0].name, "MAC")

    def test_get_all_leaf(self):
        arch = _simple_arch()
        leaves = list(arch.get_nodes_of_type(Leaf))
        self.assertEqual(len(leaves), 3)

    def test_get_memory_in_fork_arch(self):
        arch = _fork_arch()
        memories = list(arch.get_nodes_of_type(Memory))
        names = {m.name for m in memories}
        self.assertIn("MainMemory", names)
        self.assertIn("GlobalBuffer", names)
        self.assertIn("Register", names)

    def test_get_compute_in_fork_arch(self):
        arch = _fork_arch()
        computes = list(arch.get_nodes_of_type(Compute))
        names = {c.name for c in computes}
        self.assertEqual(names, {"ScalarUnit", "MAC"})

    def test_get_fanout_in_fork_arch(self):
        arch = _fork_arch()
        fanouts = list(arch.get_nodes_of_type(Fanout))
        self.assertEqual(len(fanouts), 1)
        self.assertEqual(fanouts[0].name, "ArrayFanout")


# ============================================================================
# Arch._flatten
# ============================================================================


class TestArchFlatten(unittest.TestCase):
    def test_flatten_simple(self):
        arch = _simple_arch()
        flat = arch._flatten("MAC")
        names = [n.name for n in flat]
        self.assertEqual(names, ["MainMemory", "GlobalBuffer", "MAC"])

    def test_flatten_returns_leaf_instances(self):
        arch = _simple_arch()
        flat = arch._flatten("MAC")
        for node in flat:
            self.assertIsInstance(node, Leaf)

    def test_flatten_ends_with_compute(self):
        arch = _simple_arch()
        flat = arch._flatten("MAC")
        self.assertIsInstance(flat[-1], Compute)

    def test_flatten_fork_mac_path(self):
        """Flattening to MAC should skip the Fork (ScalarUnit) branch."""
        arch = _fork_arch()
        flat = arch._flatten("MAC")
        names = [n.name for n in flat]
        self.assertIn("MainMemory", names)
        self.assertIn("GlobalBuffer", names)
        self.assertIn("MAC", names)
        # ScalarUnit is in the fork, should not be in the MAC path
        self.assertNotIn("ScalarUnit", names)

    def test_flatten_fork_scalar_path(self):
        """Flattening to ScalarUnit should go through the Fork branch."""
        arch = _fork_arch()
        flat = arch._flatten("ScalarUnit")
        names = [n.name for n in flat]
        self.assertIn("MainMemory", names)
        self.assertIn("GlobalBuffer", names)
        self.assertIn("ScalarUnit", names)
        # MAC is the main compute, should not be in ScalarUnit's path
        self.assertNotIn("MAC", names)

    def test_flatten_nonexistent_compute_returns_no_compute(self):
        """Flattening to a non-existent compute node returns a list without that node."""
        arch = _simple_arch()
        # The arch just collects non-matching leaves until it runs out of nodes.
        flat = arch._flatten("NonexistentCompute")
        names = [n.name for n in flat]
        self.assertNotIn("NonexistentCompute", names)


# ============================================================================
# Spec._get_flattened_architecture
# ============================================================================


class TestSpecGetFlattenedArchitecture(unittest.TestCase):
    def _make_spec(self, arch=None, workload=None):
        spec = Spec(
            arch=arch or _simple_arch(),
            workload=workload or _simple_workload(),
        )
        return spec._spec_eval_expressions()

    def test_returns_all_paths(self):
        spec = self._make_spec()
        result = spec._get_flattened_architecture()
        self.assertIsInstance(result, list)
        self.assertGreater(len(result), 0)

    def test_single_compute_gives_one_path(self):
        spec = self._make_spec()
        result = spec._get_flattened_architecture()
        self.assertEqual(len(result), 1)
        names = [n.name for n in result[0]]
        self.assertIn("MAC", names)

    def test_path_ends_with_compute(self):
        spec = self._make_spec()
        result = spec._get_flattened_architecture()
        for path in result:
            self.assertIsInstance(path[-1], Compute)

    def test_specific_compute_node(self):
        """When compute_node is given, returns a single flat list (not list of lists)."""
        spec = self._make_spec()
        result = spec._get_flattened_architecture(compute_node="MAC")
        # Returns a single path (flat list of Leaf nodes), not list of lists
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[-1], Leaf)
        self.assertEqual(result[-1].name, "MAC")

    def test_unevaluated_spec_raises(self):
        spec = Spec(
            arch=_simple_arch(),
            workload=_simple_workload(),
        )
        with self.assertRaises(AssertionError):
            spec._get_flattened_architecture()

    def test_fork_arch_all_paths(self):
        spec = self._make_spec(arch=_fork_arch())
        result = spec._get_flattened_architecture()
        # Should have 2 paths: one to ScalarUnit and one to MAC
        self.assertEqual(len(result), 2)
        all_compute_names = {path[-1].name for path in result}
        self.assertEqual(all_compute_names, {"ScalarUnit", "MAC"})

    def test_fork_arch_specific_mac(self):
        """When compute_node is given for a fork arch, returns a single flat path."""
        spec = self._make_spec(arch=_fork_arch())
        result = spec._get_flattened_architecture(compute_node="MAC")
        # Returns a single path, not list of lists
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[-1], Leaf)
        self.assertEqual(result[-1].name, "MAC")


# ============================================================================
# Flattening from YAML examples
# ============================================================================


class TestFlattenFromYAML(unittest.TestCase):
    def test_simple_arch_flatten(self):
        arch_path = EXAMPLES_DIR / "arches" / "simple.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "matmuls.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path, jinja_parse_data={"N_EINSUMS": 1})
        evaluated = spec._spec_eval_expressions()
        result = evaluated._get_flattened_architecture()
        self.assertEqual(len(result), 1)
        names = [n.name for n in result[0]]
        self.assertEqual(names, ["MainMemory", "GlobalBuffer", "MAC"])

    def test_tpu_arch_flatten(self):
        """TPU arch requires einsum context for expressions like len(All) == 2."""
        arch_path = EXAMPLES_DIR / "arches" / "tpu_v4i.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path)
        # Full evaluation with einsum provides needed context for expressions
        evaluated = spec._spec_eval_expressions(einsum_name="Matmul1")
        result = evaluated._get_flattened_architecture()
        # TPU has 2 computes: ScalarUnit and MAC
        self.assertEqual(len(result), 2)

    def test_tpu_mac_path_includes_register(self):
        arch_path = EXAMPLES_DIR / "arches" / "tpu_v4i.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path)
        evaluated = spec._spec_eval_expressions(einsum_name="Matmul1")
        result = evaluated._get_flattened_architecture(compute_node="MAC")
        names = [n.name for n in result]
        self.assertIn("Register", names)
        self.assertIn("ArrayFanout", names)
        self.assertEqual(names[-1], "MAC")

    def test_tpu_scalar_path_excludes_register(self):
        arch_path = EXAMPLES_DIR / "arches" / "tpu_v4i.yaml"
        wl_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not arch_path.exists() or not wl_path.exists():
            self.skipTest("YAML files not found")
        spec = Spec.from_yaml(arch_path, wl_path)
        evaluated = spec._spec_eval_expressions(einsum_name="Matmul1")
        result = evaluated._get_flattened_architecture(compute_node="ScalarUnit")
        names = [n.name for n in result]
        self.assertNotIn("Register", names)
        self.assertNotIn("ArrayFanout", names)
        self.assertEqual(names[-1], "ScalarUnit")


# ============================================================================
# Duplicate name detection
# ============================================================================


class TestDuplicateNameDetection(unittest.TestCase):
    def test_duplicate_leaf_name_rejected_by_pydantic(self):
        """Architecture with duplicate leaf names is rejected during construction."""
        # Pydantic validation catches duplicate names, preventing construction
        with self.assertRaises(Exception):
            Arch(
                nodes=[
                    Memory(
                        name="Mem",
                        size=100,
                        actions=[
                            {"name": "read", "energy": 1, "latency": 0},
                            {"name": "write", "energy": 1, "latency": 0},
                        ],
                        leak_power=0,
                        area=0,
                    ),
                    Memory(
                        name="Mem",
                        size=200,
                        actions=[
                            {"name": "read", "energy": 1, "latency": 0},
                            {"name": "write", "energy": 1, "latency": 0},
                        ],
                        leak_power=0,
                        area=0,
                    ),
                    Compute(
                        name="MAC",
                        actions=[{"name": "compute", "energy": 1, "latency": 1}],
                        leak_power=0,
                        area=0,
                    ),
                ]
            )


if __name__ == "__main__":
    unittest.main()



import matplotlib.axes as mpax
import matplotlib.pyplot as plt
from accelforge._accelerated_imports import np
from accelforge._accelerated_imports import pd

from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import PmappingDataframe


DATAFLOW_COLUMN = "dataflow"


def plot_ski_slope(
    data: pd.DataFrame,
    categorize_by_dataflow: bool = False,
    split_by_dataflow: bool = False,
    ax: mpax.Axes = None,
    **kwargs
):
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    if categorize_by_dataflow or split_by_dataflow:
        _add_dataflow_to_data(data)

    if not split_by_dataflow:
        data = PmappingDataframe(data).data

    separated_datas = []
    labels = []
    if categorize_by_dataflow or split_by_dataflow:
        for dataflow, sub_df in data.groupby(by=DATAFLOW_COLUMN):
            separated_datas.append(PmappingDataframe(sub_df).data)
            labels.append(dataflow)
    else:
        separated_datas.append(data)
        labels.append(None)

    for label, sub_df in zip(labels, separated_datas):
        ax.plot(
            *make_staircase(
                sub_df["Occupancy"].to_numpy(), sub_df["Offchip Accesses"].to_numpy()
            ),
            label=label,
            **kwargs
        )

    ax.set_xlabel("Capacity")
    ax.set_ylabel("Off-chip Accesses")
    ax.set_xscale("log")
    ax.set_yscale("log")

    return fig, ax


def make_staircase(x: np.array, y: np.array):
    """Make staircase from x, y coordinates."""
    sort_idx = np.argsort(x)
    x, y = x[sort_idx], y[sort_idx]

    shifted_x = x[1:]
    shifted_y = y[:-1]
    x = np.concat([x, shifted_x])
    y = np.concat([y, shifted_y])

    sort_idx = np.lexsort([x, -y])
    x, y = x[sort_idx], y[sort_idx]

    return x, y


def _add_dataflow_to_data(data: pd.DataFrame):
    data[DATAFLOW_COLUMN] = data["__Mappings"].apply(_dataflow_from_fullmapping)


def _dataflow_from_fullmapping(fullmapping: str):
    fullmapping = fullmapping.strip("[")
    fullmapping = fullmapping.strip("]")
    dataflow = []
    for term in fullmapping.split(","):
        if term[0] == "T":
            dataflow.append(int(term[1:].split(" size ")))
    return tuple(dataflow)


if __name__ == "__main__":
    print(make_staircase(np.array([1, 2, 3, 4]), np.array([4, 3, 2, 1])))



from sphinx.ext.autodoc import ClassDocumenter, INSTANCEATTR, ObjectMember
from sphinx.util.inspect import safe_getattr
import inspect
from pydantic import BaseModel


# Default ignore lists
DEFAULT_IGNORE = [
    'object',
    'ABC',
    'BaseModel',
    'Enum',
    'Flag',
    'IntEnum',
    'IntFlag',
    'Generic',
    'object',
    'ABC',
    'BaseModel',
    'EvalableModel',
    'EvalableList',
    'EvalableDict',
    'EvalExtras',
    'NonEvalableModel',
    '_FromYAMLAble',
    'EvalsTo',
    'Enum',
    'Flag',
    'IntEnum',
    'IntFlag',
    'Generic',
    'model_validate',
    'model_validate_json',
    'model_validate_strings',
    'parse_file',
    'parse_obj',
    'parse_raw',
    'schema',
    'schema_json',
    'model_dump',
    'model_dump_json',
    'model_copy',
    'construct',
    'copy',
    'dict',
    'json',
    'update_forward_refs',
    'model_post_init',
    'model_config',
    'model_fields',
    'model_fields_set',
    'model_extra',
    'model_private',
    'model_rebuild',
    'get_fields',
    'get_validator',
    'all_fields_default',
    'model_dump_non_none',
] + list(BaseModel.__dict__.keys())

class InheritedAttributesClassDocumenter(ClassDocumenter):
    """Enhanced ClassDocumenter that includes inherited attributes."""

    priority = ClassDocumenter.priority + 1

    def get_object_members(self, want_all):
        """Override to include inherited members."""
        # Get the normal members first
        members_check_module, members = super().get_object_members(want_all)

        # Get configuration
        config = self.env.config
        ignore = getattr(config, 'inherited_attributes_ignore', []) + DEFAULT_IGNORE

        # Collect ALL attributes defined directly on this class (not inherited)
        own_members = set()
        if hasattr(self.object, '__dict__'):
            own_members.update(self.object.__dict__.keys())
        if hasattr(self.object, 'model_fields'):
            own_members.update(self.object.model_fields.keys())
        if hasattr(self.object, '__fields__'):
            own_members.update(self.object.__fields__.keys())

        # Filter out members from ignored classes that were already included
        # BUT keep them if they're overridden in the current class
        filtered_members = []
        for member in members:
            if hasattr(member, '__getitem__') and len(member) >= 4:
                name = member[0]
                member_class = member[3] if len(member) > 3 else None

                # Keep if it's defined directly on this class
                if name in own_members:
                    filtered_members.append(member)
                    continue

                # Skip if from an ignored class (and not overridden)
                if member_class and self._should_ignore_class(member_class, ignore):
                    continue

                # Skip ignored functions/attributes (only if not overridden)
                if name in ignore:
                    continue

                filtered_members.append(member)
            else:
                filtered_members.append(member)

        members = filtered_members

        # Get MRO
        try:
            mro = inspect.getmro(self.object)
        except (AttributeError, TypeError):
            return members_check_module, members

        # Track seen names - includes both members from super() and own_members
        # ObjectMember is a namedtuple, access by index
        seen_names = set()
        for member in members:
            if hasattr(member, '__getitem__'):
                seen_names.add(member[0])
        # Add all own members to seen_names to prevent inheritance
        seen_names.update(own_members)

        # Collect inherited members
        inherited = []

        for parent_class in mro[1:]:  # Skip self
            # Check if we should stop FIRST, before processing any members
            if self._should_ignore_class(parent_class, ignore):
                break

            # Get members from parent
            for name in dir(parent_class):
                # Skip underscore-prefixed
                if name.startswith('_'):
                    continue

                # Skip if already seen (including own_members)
                if name in seen_names:
                    continue

                # Skip ignored
                if name in ignore:
                    continue

                try:
                    obj = safe_getattr(parent_class, name)

                    # Skip callables in ignore list
                    if callable(obj) and name in ignore:
                        continue

                    # Create ObjectMember
                    inherited.append(ObjectMember(name, obj))
                    seen_names.add(name)

                except (AttributeError, TypeError):
                    continue

            # Check Pydantic fields
            if hasattr(parent_class, 'model_fields'):
                for field_name in parent_class.model_fields:
                    if field_name.startswith('_'):
                        continue
                    # Skip if already seen (including own_members)
                    if field_name in seen_names:
                        continue
                    if field_name in ignore:
                        continue

                    inherited.append(ObjectMember(field_name, INSTANCEATTR))
                    seen_names.add(field_name)

            elif hasattr(parent_class, '__fields__'):
                for field_name in parent_class.__fields__:
                    if field_name.startswith('_'):
                        continue
                    # Skip if already seen (including own_members)
                    if field_name in seen_names:
                        continue
                    if field_name in ignore:
                        continue

                    inherited.append(ObjectMember(field_name, INSTANCEATTR))
                    seen_names.add(field_name)

        # Combine
        all_members = list(members) + inherited

        return members_check_module, all_members

    def _should_ignore_class(self, cls, ignore_classes):
        """Check if a class should be ignored."""
        name = getattr(cls, '__name__', '')
        module = getattr(cls, '__module__', '')

        # Ignore if not part of accelforge package
        if module and not module.startswith('accelforge'):
            return True

        # Check against ignore list using just the class name
        for ignore_pattern in ignore_classes:
            if name == ignore_pattern:
                return True

        return False

def setup(app):
    """Setup the extension."""
    # Only add config value if it doesn't exist
    if not hasattr(app.config, 'inherited_attributes_ignore'):
        app.add_config_value('inherited_attributes_ignore', [], 'env')

    app.add_autodocumenter(InheritedAttributesClassDocumenter, override=True)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""
Handles the ISL temporal reuse functions.
"""

from dataclasses import dataclass

import islpy as isl

from accelforge.model._looptree.reuse.isl.isl_functions import map_to_shifted
from accelforge.model._looptree.reuse.isl.mapping_to_isl.types import (
    TEMPORAL_TAGS,
    Fill,
    Occupancy,
    Tag,
)


@dataclass(frozen=True)
class TemporalReuse:
    """Results for an temporal reuse analysis."""

    effective_occupancy: Occupancy
    """TODO: Figure this out."""
    fill: Fill
    """Data deliveries to locations in spacetime that need to be made."""


def analyze_temporal_reuse(
    occ: Occupancy, exploit_reuse: bool = True, multi_loop_reuse: bool = True
) -> TemporalReuse:
    """
    Computes the required fill to satisfy the buffer occupancy.
    If the buffer can `exploit_reuse`, then the fill will only consist
    of data not currently resident in buffer.

    Parameters
    ----------
    occ:
        The logical occupancy to be temporally analyzed.
    exploit_reuse:
        Temporally exploits reuse through persisting data currently in buffer
        to the next time step.
    multi_loop_reuse:
        Whether when this loop, or one above it in the memory hierarchy, loops,
        does the buffer flush.

    Returns
    -------
    A struct containing a `..types.Fill` which is how to load data into the buffer
    across time and a `..types.Occupancy` describing the effective_occupancy across
    time (i.e., what data needs to be persisted in the buffer per time step and what
    can be ignored/purged).

    TODO: Make sure spaces are named properly
    """
    if exploit_reuse:
        return fill_from_occupancy(occ, multi_loop_reuse)
    return TemporalReuse(occ, Fill(occ.tags, occ.map_))


def fill_from_occupancy(
    occupancy: Occupancy, multiple_loop_reuse: bool
) -> TemporalReuse:
    """
    Given an occupancy and if you're allowed to reuse across loops, calculate the
    `fill` and the `effective_occupancy` per time step.

    Parameters
    ----------
    occupancy:
        The logical occupancy of data in logical buffers.
    multi_loop_reuse:
        If you are allowed to use data between loop iterations.

    Returns
    -------
    A `TemporalReuse` object that contains the `fill` and `effective_occpancy`
    of the lowest buffer level.
    """
    # Iterates through each dimension in reverse order (i.e., deepest loop first)
    occ = occupancy.map_.copy()
    tags = occupancy.tags.copy()
    for dim_idx, tag in reversed(list(enumerate(occupancy.tags))):
        if not isinstance(tag, TEMPORAL_TAGS):
            continue
        # Check if temporal dimension is "trivial," i.e., equals a singular value
        proj_occ: isl.Map = occ.project_out(
            isl.dim_type.in_, dim_idx, 1
        ).set_tuple_name(
            isl.dim_type.in_, f"{occ.get_tuple_name(isl.dim_type.in_)}_abridged"
        )
        reinserted_occ: isl.Map = (
            proj_occ.insert_dims(isl.dim_type.in_, dim_idx, 1).set_tuple_name(
                isl.dim_type.in_,
                occ.get_tuple_name(isl.dim_type.in_).removesuffix("_abridged"),
            )
        ).intersect_domain(occ.domain())

        if occ.plain_is_equal(reinserted_occ) or occ.is_equal(reinserted_occ):
            occ = proj_occ
            tags.pop(dim_idx)
            continue

        # Nontrivial analysis
        time_shift: isl.Map
        if not multiple_loop_reuse:
            # TODO: Verify space names are preserved and/or replace.
            time_shift = map_to_shifted(occ.domain().get_space(), dim_idx, -1)
        # Calculates the time_shift assuming no cache flushing for loops.
        else:
            # TODO: this is a better way of getting time_shift. Use method to
            # replace the other branch (!multi_loop_reuse)
            time_shift = construct_time_shift(occ, tags)

        # Gets the fill (i.e., feeds data not currently in buffer).
        occ_before: isl.Map = time_shift.apply_range(occ)
        fill: isl.Map = occ.subtract(occ_before)

        return TemporalReuse(Occupancy(tags, occ), Fill(tags, fill))

    return TemporalReuse(Occupancy(tags, occ.coalesce()), Fill(tags, occ.coalesce()))


def construct_time_shift(occ: isl.Map, tags: list[Tag]):
    """
    Given an occupancy and its input dimension tags, create the proper spatial
    and temporal separation objects.

    Parameters
    ----------
    occ:
        The occupancy map we're analyzing the reuse for.
    tags:
        The tags of what an input represents.

    Returns
    -------
    time_shift:
        Relation of the current time step to the previous one across loops.
    """
    # Creates the spacetime deconstruction to the two separate components.
    spacetime: isl.Set = occ.domain()
    spacetime_to_time: isl.Map = isl.Map.identity(spacetime.get_space().map_from_set())
    spacetime_to_space: isl.Map = isl.Map.identity(spacetime.get_space().map_from_set())
    # Prunes out the output dimensions that do not correspond to the
    # correct mapping into a generic space-to-space relation.
    for idx, t in reversed(list(enumerate(tags))):
        if not isinstance(t, TEMPORAL_TAGS):
            spacetime_to_time = spacetime_to_time.project_out(isl.dim_type.out, idx, 1)
        else:
            spacetime_to_space = spacetime_to_space.project_out(
                isl.dim_type.out, idx, 1
            )
    # Gets the names correct after transformations.
    spacetime_to_time = spacetime_to_time.set_tuple_name(
        isl.dim_type.out, f"{spacetime.get_tuple_name()}_time"
    )
    spacetime_to_space = spacetime_to_space.set_tuple_name(
        isl.dim_type.out, f"{spacetime.get_tuple_name()}_space"
    )

    # Properly constrains the spacetime_to_time's domain.
    spacetime_to_time = spacetime_to_time.intersect_domain(spacetime)
    time_: isl.Set = spacetime_to_time.range()
    # Creates a map of time_ to previous regions of time_.
    time_to_past: isl.Map = (
        isl.Map.lex_gt(time_.get_space()).intersect_domain(time_).intersect_range(time_)
    )
    # Restricts the relation to only the most recent previous region of time_.
    time_to_most_recent_past = time_to_past.lexmax()
    # Relates the current spacetime to its direct predecessor in time.
    time_shift: isl.Map = spacetime_to_time.apply_range(
        time_to_most_recent_past.apply_range(spacetime_to_time.reverse())
    )

    # Prunes spatial relations to only ones that are valid.
    spacetime_space_preserver: isl.Map = spacetime_to_space.apply_range(
        spacetime_to_space.reverse()
    )
    # Intersects with time_shift as space information is lost with the compression of
    # spacetime to time_ and then rexpansion to past time_.
    return time_shift.intersect(spacetime_space_preserver)



from accelforge.frontend.mapping import Reservation, Loop, Mapping


def get_fused_loops_per_tensor(
    pmapping: Mapping, intermediate_tensors, non_fused_memory
):
    """
    Returns a dictionary mapping tensor to number of fused loops or None
    if unfused (backed in non_fused_memory).
    """
    tensor_to_n_fused_loops = {}
    n_loops = 0
    for node in pmapping.nodes:
        if isinstance(node, Reservation):
            tensor = node.tensor
            if tensor not in intermediate_tensors or tensor in tensor_to_n_fused_loops:
                continue
            if node.component == non_fused_memory:
                tensor_to_n_fused_loops[tensor] = None
            else:
                tensor_to_n_fused_loops[tensor] = n_loops
        elif isinstance(node, Loop):
            n_loops += 1
    return tensor_to_n_fused_loops



from math import isclose
import unittest
from pathlib import Path

from accelforge.frontend import Spec
from accelforge.frontend.arch import Arch
from accelforge.frontend.mapping import Mapping
from accelforge.frontend.workload import Workload

from accelforge.frontend.mapper.metrics import Metrics
from accelforge.model._looptree.accesses import (
    isl_buffer_accesses_from_buffet_actions,
    Accesses,
)
from accelforge.model._looptree.energy import gather_actions
from accelforge.model._looptree.latency import get_latency
from accelforge.model._looptree.types import Buffet
from accelforge.model._looptree.reuse.symbolic import (
    BuffetStats,
    analyze_reuse_and_add_reservations_to_mapping,
    Compute,
)
from accelforge.mapper.FFM import make_pmappings, join_pmappings


PARENT_DIR = Path(__file__).parent


class TestProcessingStage(unittest.TestCase):
    def test_toll(self):
        spec = Spec.from_yaml(
            [
                Path(__file__).parent / "toll.arch.yaml",
                Path(__file__).parent / "matmul.workload.yaml",
            ]
        )
        spec.mapper.metrics = Metrics.ENERGY
        pmappings = make_pmappings(spec)
        mappings = join_pmappings(pmappings)

        energy = mappings.data.iloc[0]["Total_energy"]
        self.assertAlmostEqual(energy, 100 * (128 * 64 * 2 + 128 * 128))


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path

import time

from accelforge.frontend.spec import Spec, Mapping
from accelforge.mapper.FFM._make_pmappings.contraints.constraints import (
    MappingConstraints,
)
from accelforge.mapper.FFM._make_pmappings.pmapper_job import Job
from accelforge.mapper.FFM._make_pmappings.make_pmappings import make_tile_shapes
from accelforge.frontend.mapper.metrics import Metrics
from accelforge.frontend import arch


class TestTileShapeExploration(unittest.TestCase):
    def test_pmapping(self):
        PARENT_DIR = Path(__file__).parent
        spec = Spec.from_yaml(
            PARENT_DIR / "conv.workload.yaml", PARENT_DIR / "four_level.arch.yaml"
        )
        spec = spec.calculate_component_area_energy_latency_leak(area=False)

        mapping = Mapping.from_yaml(PARENT_DIR / "conv_sym.mapping.yaml")

        flattened_arch = spec._get_flattened_architecture()
        memories_track_all = [
            m.name for m in flattened_arch if isinstance(m, arch.Memory)
        ]
        memories_track_pmappings_only = []

        job = Job(
            mapping=mapping,
            constraints=MappingConstraints(),
            spec=spec,
            metrics=Metrics.LATENCY,
            job_id=0,
            rank_variable_bounds={},
            memories_track_all=memories_track_all,
            memories_track_pmappings_only=memories_track_pmappings_only,
        )

        result, _ = make_tile_shapes(job)
        data, n_total_pmappings = result
        self.assertTrue("Total_latency" in data.columns)


if __name__ == "__main__":
    unittest.main()



"""
Contains shared classes of analysis.
"""

from dataclasses import dataclass
from typing import TypeAlias

from accelforge.frontend.mapping import TensorName
from accelforge.frontend.workload import EinsumName

ComponentName: TypeAlias = str


@dataclass(eq=True, frozen=True)
class Buffet:
    """
    A logical buffer that stores a tensor, an einsum operating on it, and the
    level the buffer exists on in hardware.
    """

    tensor: TensorName
    "The tensor held by the buffet."
    einsum: EinsumName
    "An einsum operating on the tensor."
    level: ComponentName
    "The abstract hardware level the buffet resides in."



from typing import overload
from sympy import Piecewise

# from accelforge.model._looptree._isl.singular import get_value_from_singular_qpolynomial
from accelforge.frontend.arch import Compute
from accelforge.model._looptree.latency.processors import LATENCY_PROCESSORS
from accelforge.model._looptree.reuse.isl import IslReuseAnalysisOutput
from accelforge.model._looptree.reuse import SymbolicAnalysisOutput

from accelforge.util._sympy.broadcast_max import Max

# from bindings._looptree import SpatialTag


def get_latency(looptree_results, mapping, workload, flattened_arch):
    comp_latency = calculate_compute_latency(looptree_results, mapping, workload)
    mem_latency = memory_latency(looptree_results, flattened_arch, mapping, workload)

    overall_latency = Max(comp_latency, *mem_latency.values())
    return overall_latency, comp_latency, mem_latency


@overload
def calculate_compute_latency(
    reuse_analysis_results: IslReuseAnalysisOutput, mapping, workload
):
    pass


@overload
def calculate_compute_latency(
    reuse_analysis_results: SymbolicAnalysisOutput, mapping, workload
):
    pass


def calculate_compute_latency(reuse_analysis_results, mapping, workload):
    if isinstance(reuse_analysis_results, IslReuseAnalysisOutput):
        return compute_isl_latency(
            reuse_analysis_results.temporal_steps, mapping, workload
        )
    elif isinstance(reuse_analysis_results, SymbolicAnalysisOutput):
        return compute_summarized_latency(
            reuse_analysis_results.compute_stats, mapping, workload
        )


def compute_summarized_latency(compute_stats, mapping, workload):
    # TODO: this is only for single-Einsum!!!
    longest_compute_latency = 0
    for stats in compute_stats.values():
        if longest_compute_latency == 0:
            longest_compute_latency = stats.max_latency
        else:
            longest_compute_latency = Max(longest_compute_latency, stats.max_latency)
    return longest_compute_latency


def compute_isl_latency(temporal_steps, mapping, workload):
    raise NotImplementedError()
    return get_value_from_singular_qpolynomial(
        _compute_latency(mapping.nodes, 0, temporal_steps, workload)[1]
    ).to_python()


def _compute_latency(mapping, top_idx: int, temporal_steps, workload):
    raise NotImplementedError()
    einsum_name_to_id = workload.einsum_name_to_id()

    next_top_idx = top_idx
    for node in mapping:
        next_top_idx += 1

        if node["type"] in LATENCY_PROCESSORS.keys():
            children_latencies = [
                _compute_latency(branch, next_top_idx, temporal_steps, workload)
                for branch in node["branches"]
            ]

            return LATENCY_PROCESSORS[node["type"]](top_idx, children_latencies)
        elif node["type"] == "compute":
            einsum = node["einsum"]
            if "incomplete" in node and node["incomplete"]:
                return ([], 0)
            einsum_id = einsum_name_to_id[einsum]
            return temporal_steps[einsum_id]


def ops_to_latency(dims, map):
    raise NotImplementedError()
    mask = [False] * len(dims)
    new_dims = []
    for i, d in enumerate(dims):
        if d == SpatialTag:
            mask[i] = True
        else:
            new_dims.append(d)
    return map.domain().identity().card()



"""Setup.py to ensure accelforge._version_scheme is importable during build."""

import sys
import os
from pathlib import Path

current_dir = Path(__file__).parent.absolute()
parent_dir = current_dir.parent

for path in [str(current_dir), str(parent_dir)]:
    if path not in sys.path:
        sys.path.insert(0, path)

pythonpath = os.environ.get("PYTHONPATH", "")
if pythonpath:
    for path in pythonpath.split(os.pathsep):
        if path and path not in sys.path:
            sys.path.insert(0, path)

import importlib.util
_version_scheme_path = current_dir / "accelforge" / "_version_scheme.py"
spec = importlib.util.spec_from_file_location("_version_scheme", _version_scheme_path)
_version_scheme = importlib.util.module_from_spec(spec)
spec.loader.exec_module(_version_scheme)
post_version = _version_scheme.post_version
no_local = _version_scheme.no_local

from setuptools import setup

setup(
    use_scm_version={
        "version_scheme": post_version,
        "local_scheme": no_local,
        "write_to": "accelforge/_version.py",
        "fallback_version": "1.0",
    }
)



from collections import defaultdict
import itertools

from accelforge.frontend import arch
from accelforge.frontend.spec import Spec
from accelforge.mapper.FFM._join_pmappings.join_pmappings import PmappingGroup
from accelforge.mapper.FFM._join_pmappings.compatibility import Loop, Compatibility
from accelforge.util._frozenset import fzs
from accelforge.mapper.FFM._join_pmappings.join_pmappings import (
    make_full_equivalent_rank_variables,
)


class MapspaceGlobals:
    def __init__(
        self,
        pmapping_groups: dict[str, list[PmappingGroup]],
        spec: Spec,
        objective_function_cols: list[str] = None,
        flattened_architecture: list[arch.Leaf] = None,
    ):
        self.pmapping_groups = pmapping_groups
        self.einsum_names = spec.workload.einsum_names
        self.einsum2ranks = {
            einsum_name: spec.workload.einsums[einsum_name].rank_variables
            for einsum_name in self.einsum_names
        }
        self.einsum2tensors = {
            einsum_name: spec.workload.einsums[einsum_name].tensor_names
            for einsum_name in self.einsum_names
        }
        self.tensor_names = set().union(
            *(self.einsum2tensors[e] for e in self.einsum_names)
        )
        self.tensor_names_used_in_multiple_einsums = (
            spec.workload.tensor_names_used_in_multiple_einsums
        )
        self.pairwise_equivalent_ranks = (
            spec.workload.get_pairwise_equivalent_rank_variables()
        )
        self.full_equivalent_ranks = make_full_equivalent_rank_variables(
            self.pairwise_equivalent_ranks
        )

        self.resource2capacity = {}
        flattened_architecture = (
            flattened_architecture or spec.get_flattened_architecture()
        )
        for l in flattened_architecture:
            if isinstance(l, arch.Memory):
                self.resource2capacity[l.name] = l.size

        self.objective_function_cols = objective_function_cols
        self.rank_translations = self._create_rank_translations(self.einsum2ranks)

        for i, (left_id, left_sims) in enumerate(pmapping_groups.items()):
            for j, (right_id, right_sims) in enumerate(pmapping_groups.items()):
                if i >= j:
                    continue

                left_live = self.get_live_tensors(*self.einsum_names[: i + 1])
                right_live = self.get_live_tensors(*self.einsum_names[j:])
                left_tensors = self.get_tensors(self.einsum_names[i])
                right_tensors = self.get_tensors(self.einsum_names[j])

                if not (left_live & right_live):
                    continue
                print(f"Checking {left_id} {right_id}")

                right_tilings = {
                    s.compatibility.clear_dead_tensors(
                        live_tensors=left_live
                    ).clear_dead_tensors(left_tensors, keep_loops=True)
                    for s in right_sims
                }
                assert right_tilings, f"R {left_id} {right_id}"
                for s in list(left_sims):
                    for t in self.get_possible_translations(s.compatibility, right_id):
                        t = t.clear_dead_tensors(live_tensors=right_live)
                        t = t.clear_dead_tensors(
                            live_tensors=right_tensors, keep_loops=True
                        )
                        if t in right_tilings:
                            break
                    else:
                        left_sims.remove(s)
                assert (
                    left_sims
                ), f"Removed all of left {left_id} while checking right {right_id}"

                left_tilings = {
                    s.compatibility.clear_dead_tensors(
                        live_tensors=right_live
                    ).clear_dead_tensors(right_tensors, keep_loops=True)
                    for s in left_sims
                }
                assert left_tilings, f"L {left_id} {right_id}"
                for s in list(right_sims):
                    for t in self.get_possible_translations(s.compatibility, left_id):
                        t = t.clear_dead_tensors(live_tensors=left_live)
                        t = t.clear_dead_tensors(
                            live_tensors=left_tensors, keep_loops=True
                        )
                        if t in left_tilings:
                            break
                    else:
                        right_sims.remove(s)
                assert (
                    right_sims
                ), f"Removed all of right {right_id} while checking left {left_id}"

        self.tensor2possible_loops_above = self._create_tensor2possible_loops_above()
        self.tensor2possible_loops_above_set = {
            k: {k2: set(v2) for k2, v2 in v.items()}
            for k, v in self.tensor2possible_loops_above.items()
        }
        self.tensor2memories = self._create_tensor2memories()
        self.einsum_tiling_2_sim = self._create_einsum_tiling_2_sim()
        self.einsum_rank_index_to_loops = self._create_einsum_rank_index_to_loops()
        (
            self.compatibility2leftcompatibility,
            self.compatibility2rightcompatibility,
            self.leftcompatibility2tiling,
            self.rightcompatibility2tiling,
        ) = self._create_compatibility()
        self.size_scale = len(self.einsum2ranks)
        n_optimal = sum(
            len(s.mappings.data)
            for simlist in self.pmapping_groups.values()
            for s in simlist
        )
        n_pmappings = sum(
            s.mappings.n_pmappings
            for simlist in self.pmapping_groups.values()
            for s in simlist
        )
        self.find_pmapping_scale = n_pmappings / n_optimal
        self.aliased_tensors = spec.workload.get_tensor_copies()

    def get_live_tensors(self, *einsums: str):
        return set.union(*(self.einsum2tensors[e] for e in einsums))

    def _create_compatibility(self):
        tiling2leftcompatibility = {}
        tiling2rightcompatibility = {}

        def tilings2compatibility(tilings: list[Compatibility], live_tensors: set[str]):
            return {t: t.clear_dead_tensors(live_tensors=live_tensors) for t in tilings}

        for i, (einsum_name, pm_group_list) in enumerate(self.pmapping_groups.items()):
            if i > 0:
                prev_live = self.get_live_tensors(*self.einsum_names[:i])
                tiling2leftcompatibility[einsum_name] = tilings2compatibility(
                    [s.compatibility for s in pm_group_list],
                    prev_live,
                )
            if i < len(self.pmapping_groups) - 1:
                next_live = self.get_live_tensors(*self.einsum_names[i + 1 :])
                tiling2rightcompatibility[einsum_name] = tilings2compatibility(
                    [s.compatibility for s in pm_group_list],
                    next_live,
                )

        leftcompatibility2tiling = {}
        rightcompatibility2tiling = {}
        for einsum_name in self.einsum_names:
            for src, dst in (
                (tiling2leftcompatibility, leftcompatibility2tiling),
                (tiling2rightcompatibility, rightcompatibility2tiling),
            ):
                if einsum_name not in src:
                    continue
                dst = dst.setdefault(einsum_name, {})
                for k, v in src[einsum_name].items():
                    dst.setdefault(v, []).append(k)
        return (
            tiling2leftcompatibility,
            tiling2rightcompatibility,
            leftcompatibility2tiling,
            rightcompatibility2tiling,
        )

    def _create_einsum_tiling_2_sim(self):
        einsum_tiling_2_sim = {}
        for e, pm_group_list in self.pmapping_groups.items():
            cur_sims = defaultdict(list)
            for sim in pm_group_list:
                cur_sims[sim.compatibility].append(sim)
            einsum_tiling_2_sim[e] = {}
            for t, s in cur_sims.items():
                s = PmappingGroup.concat(s)
                einsum_tiling_2_sim[e][t] = s
        return einsum_tiling_2_sim

    def _create_tensor2possible_loops_above(self):
        tensor2possible_loops_above = {}
        for einsum_name, pm_group_list in self.pmapping_groups.items():
            tensor2possible_loops_above[einsum_name] = defaultdict(set)
            for sim in pm_group_list:
                for tensor in sim.compatibility.tensors:
                    tensor2possible_loops_above[einsum_name][tensor] |= set(
                        sim.compatibility.loops[: tensor.above_loop_index]
                    )
        return {
            e: {s: list(l) for s, l in d.items()}
            for e, d in tensor2possible_loops_above.items()
        }

    def _create_tensor2memories(self):
        tensor2memories = {}
        for t in self.tensor_names_used_in_multiple_einsums:
            possible_memories = []
            for einsum_name, pm_group_list in self.pmapping_groups.items():
                cur_memories = set()
                if t not in pm_group_list[0].tensor_names:
                    continue
                for sim in pm_group_list:
                    tensor = sim.compatibility.get_tensor_by_name(t)
                    cur_memories.add(tensor)
                possible_memories.append(cur_memories)
            if possible_memories:
                tensor2memories[t] = list(set.intersection(*possible_memories))
            else:
                raise ValueError(f"No memories for {t}")
        return tensor2memories

    def _create_rank_translations(self, einsum2ranks: dict[str, set[str]]):
        rank_translations = {}
        for einsum_name, ranks in einsum2ranks.items():
            translations = {einsum_name2: {} for einsum_name2 in self.einsum_names}
            for einsum_name2, ranks2 in einsum2ranks.items():
                for rank in ranks:
                    equiv = self.full_equivalent_ranks[rank] & ranks2
                    translations[einsum_name2][rank] = equiv
            rank_translations[einsum_name] = {
                k: {k2: list(v2) for k2, v2 in v.items()}
                for k, v in translations.items()
            }
        return rank_translations

    def _create_full_equivalent_ranks(
        self, pairwise_equivalent_ranks: dict[str, set[str]]
    ):
        full_equivalent_ranks = {
            k: set(v) for k, v in pairwise_equivalent_ranks.items()
        }
        changed = True
        while changed:
            changed = False
            for r in full_equivalent_ranks:
                for r2 in list(full_equivalent_ranks[r]):
                    for r3 in list(full_equivalent_ranks[r2]):
                        if r3 in full_equivalent_ranks[r]:
                            continue
                        changed = True
                        full_equivalent_ranks[r].add(r3)
        return full_equivalent_ranks

    def _create_einsum_rank_index_to_loops(
        self,
    ) -> dict[str, dict[str, dict[int, list[Loop]]]]:
        einsum_rank_index_to_loops = {}
        for einsum_name, pm_group_list in self.pmapping_groups.items():
            einsum_rank_index_to_loops[einsum_name] = {}
            for sim in pm_group_list:
                for rank_index, loop in enumerate(sim.compatibility.loops):
                    x = einsum_rank_index_to_loops[einsum_name].setdefault(
                        loop.rank_variable_name, {}
                    )
                    x.setdefault(rank_index, []).append(loop)
        return einsum_rank_index_to_loops

    def get_tensors(self, *einsums: str):
        return set.union(*(self.einsum2tensors[e] for e in einsums))

    def get_possible_translations(self, t: Compatibility, to_einsum: str):
        pairwise_equivalent_ranks = self.pairwise_equivalent_ranks
        full_equivalent_ranks = self.full_equivalent_ranks
        right_ranks = self.einsum2ranks[to_einsum]

        def translate_loop(l: Loop):
            compatible_ranks = (
                set.union(*(full_equivalent_ranks[n] for n in l.rank_variable_names))
                & right_ranks
            )
            pairwise_compatible_ranks = (
                set.union(
                    *(pairwise_equivalent_ranks[n] for n in l.rank_variable_names)
                )
                & right_ranks
            )
            if len(pairwise_compatible_ranks) > 1:
                return
            for n in compatible_ranks:
                yield Loop(fzs((n,)), l.bound, l.is_spatial)

        for loops in itertools.product(*map(translate_loop, t.loops)):
            yield Compatibility(loops, t.tensors, t.tags)
